from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.IType import IType
from belenios.models.BaseModel import BaseModel


class EmbeddingModel(BaseModel):
    padding = Column(IType, nullable=False)
    bits_per_int = Column(IType, nullable=False)
    group_id: int = Column(Integer, ForeignKey('group_model.id'), nullable=False)
    group: Mapped['GroupModel'] = relationship('GroupModel', back_populates='embedding')

    def __str__(self):
        return (
            f"EmbeddingModel(\n"
            f"    padding={self.padding},\n"
            f"    bits_per_int={self.bits_per_int},\n"
            # f"    group_id={self.group_id},\n"
            # f"    group={self.group}\n"
            f")"
        )


from abc import abstractmethod

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class AnswerModel(BaseModel):
    # __abstract__ = True
    type = Column(StringType, nullable=False)
    ballot_id = Column(Integer, ForeignKey('ballot_model.id'), nullable=False)
    __mapper_args__ = {
        "polymorphic_identity": "AnswerModel",
        "polymorphic_on": "type",
    }

    @abstractmethod
    def to_hash_sha256(self):
        pass



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.StringType import StringType
from belenios.custom_types.belenios.ArrayOfStringsType import ArrayOfStringsType
from belenios.models.BaseModel import BaseModel
from belenios.utilities.Utility import Utility


class QuestionModel(BaseModel):
    # __abstract__ = True
    type = Column(StringType, nullable=False)
    question = Column(StringType, nullable=True)
    answers = Column(ArrayOfStringsType, nullable=True)
    extra = Column(StringType, nullable=True)
    election_id: int = Column(Integer, ForeignKey('election_model.id'), nullable=False)
    election: Mapped['ElectionModel'] = relationship('ElectionModel', back_populates='questions')
    __mapper_args__ = {
        "polymorphic_identity": "QuestionModel",
        "polymorphic_on": "type",
    }
    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"type={self.type}" + delimiter +
                f"question={self.question}" + delimiter +
                f"answers={self.answers}" + delimiter +
                f"extra={self.extra}" + delimiter
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash

    def __str__(self):
        return (
            f"QuestionModel(\n"
            f"    type={self.type},\n"
            f"    question={self.question},\n"
            f"    answers={self.answers},\n"
            f"    extra={self.extra}\n"
            f")"
        )


import json

from belenios.config.config import AppConfig
from belenios.core.belenios.ServerCore import ServerCore
from belenios.custom_types.belenios.EventType import EventType
from belenios.models.belenios.ElectionBundleModel import ElectionBundleModel
from belenios.models.belenios.ElectionModel import ElectionModel
from belenios.models.belenios.GroupModel import GroupModel
from belenios.models.belenios.QuestionHModel import QuestionHModel
from belenios.models.belenios.SetupDataModel import SetupDataModel
from belenios.models.belenios.SizedEncryptedTallyModel import SizedEncryptedTallyModel
from belenios.models.belenios.TallyModel import TallyModel
from belenios.models.belenios.VoterModel import VoterModel
from belenios.utilities.Utility import Utility
from belenios.utilities.command_line.AppObject import AppObject
from belenios.utilities.command_line.DbSession import DbSession


class ServerAdministratorCore:
    @staticmethod
    def start_new_election(name, description, administrator, credential_authority, group_name):
        # 1. A starts the preparation of an election,
        election_bundle = ElectionBundleModel()
        election_bundle.setup_data = SetupDataModel()
        # providing in particular the questions and the list of voters
        election = ElectionModel()
        election.version = 1
        election.name = name
        election.description = description
        election.public_key = None
        election.administrator = administrator
        election.credential_authority = credential_authority
        group = GroupModel()
        if AppConfig().el_gamal_groups.get(group_name, None) is None:
            AppObject().print("Group invalid")
            return None
        group.description = AppConfig().el_gamal_groups.get(group_name, None).description
        group.p = AppConfig().el_gamal_groups.get(group_name).p
        group.g = AppConfig().el_gamal_groups.get(group_name).g
        group.q = AppConfig().el_gamal_groups.get(group_name).q
        election.group = group
        election_bundle.election = election
        election_bundle.tally = TallyModel()
        election_bundle.tally.sized_encrypted_tally = SizedEncryptedTallyModel()
        DbSession().session.add(election_bundle)
        DbSession().session.commit()
        return election_bundle

    @staticmethod
    def add_question(election, question_text, answers, min, max, blank, ):
        question = QuestionHModel()
        question.question = question_text
        # question.answers = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        question.answers = answers
        question.min = min
        question.max = max
        question.blank = blank
        election.questions.append(question)
        DbSession().session.commit()
        return question

    @staticmethod
    def add_voter(election_uuid, email, weight):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None, None

        # Check if a voter with the same email already exists
        existing_voter = DbSession().session.query(VoterModel).filter_by(email=email).first()
        if existing_voter:
            # Merge the existing voter with the election bundle
            election_bundle.voters.append(existing_voter)
            existing_voter.weight = weight  # Update the weight if necessary
            DbSession().session.commit()
            return existing_voter
        else:
            # Create a new voter
            new_voter = VoterModel(email=email, weight=weight)
            election_bundle.voters.append(new_voter)
            DbSession().session.commit()
            return new_voter

    # @staticmethod
    # def add_single_trustees(election_uuid, count):
    #     election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
    #     if election_bundle is None:
    #         print("Election uuid invalid")
    #         return None, None
    #     ServerCore().generates_trustee_public_key(election_uuid, count)
    #     # 2. S checks γ
    #     # We can check is the key and the pok is correctly generated :
    #     if TrusteesKeysProtocol().check_proof_trustee_key(election_bundle.election,
    #                                                       single_trustee.trustee_public_key):
    #         print("Trustee key is valid")
    #     else:
    #         print("Trustee key is invalid")
    #         return False
    #     return

    @staticmethod
    def check_perdersen_trustee_voutput(election_uuid, trustee_id, voutput):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        pedersen_trustee = None
        for pedersen_trustee_iter in election_bundle.trustee_kinds:
            if pedersen_trustee_iter.trustee_kind_type == "PedersenTrusteeModel":
                pedersen_trustee = pedersen_trustee_iter
                break
        if pedersen_trustee is None:
            print("Invalid trustee id")
            return None
        prod = 1
        for i in range(len(pedersen_trustee.certs)):
            A_i = json.loads(election_bundle.polynomials[i].coefexps.signed_object.value)
            # print(A_i)
            for k in range(pedersen_trustee.threshold):
                jk = pow(trustee_id, k)
                A_ik = A_i[k]
                # print(A_ik)
                prod *= pow(A_ik, jk, election_bundle.election.group.p)
                prod = prod % election_bundle.election.group.p
        if voutput.trustee_public_key.public_key != prod:
            print("Invalid public key")
            return False
        return True


    @staticmethod
    def end_ballot(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        event = Utility.put_new_event(election_bundle)
        event.type = EventType.EndBallots
        DbSession().session.commit()
        return event
    @staticmethod
    def make_server_make_encrypted_tally(election_uuid):
        return ServerCore.compute_encrypted_tally(election_uuid)
    @staticmethod
    def make_server_compute_result(election_uuid):
        return ServerCore.compute_encrypted_tally(election_uuid)


from cement.utils import fs
from sqlalchemy.orm import sessionmaker

from belenios.config.config import AppConfig
from belenios.models.BaseModel import BaseModel
from belenios.utilities.DatabaseInit import DatabaseInit
from belenios.utilities.command_line.AppObject import AppObject

from sqlalchemy import create_engine, MetaData, Table, text


class DbSession:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._initialize()
        return cls._instance

    def _initialize(self):
        self.session = self._connect_sqlalchemy()

    def _connect_sqlalchemy(self):
        engine = self._get_engine()

        # Generate model to create there tables
        # from belenios.models.belenios.SignedMsgModel import SignedMsgModel
        # from belenios.models.belenios.CertModel import CertModel
        # from belenios.models.belenios.CertKeysModel import CertKeysModel
        # from belenios.models.belenios.ChannelMsgModel import ChannelMsgModel
        # from belenios.models.belenios.EncryptedMsgModel import EncryptedMsgModel
        # from belenios.models.belenios.EventModel import EventModel
        # from belenios.models.belenios.HeaderModel import HeaderModel
        # from belenios.models.belenios.TrusteePublicKeyModel import TrusteePublicKeyModel

        # Create tables
        BaseModel.metadata.create_all(engine)

        # Create a session
        Session = sessionmaker(bind=engine)
        session = Session()

        DatabaseInit().run(session)
        return session

    def _get_engine(self):
        from belenios.main import BeleniosTest
        if isinstance(AppObject().app, BeleniosTest):
            db_file = fs.abspath(AppConfig().get_test_db_file())
        else:
            db_file = fs.abspath(AppConfig().get_prod_db_file())
        # print('SQLite database file is: %s' % db_file)

        # Create engine
        if db_file == '/default/absolute/path/prod.db' or db_file == '/default/absolute/path/test.db':
            raise Exception("db_file in the class AppConfig from the file config.py is not customized")
        return create_engine('sqlite:///' + db_file)

    def drop_all_tables(self):
        engine = self._get_engine()

        # Reflect the existing database schema
        metadata = MetaData()
        metadata.reflect(bind=engine)

        # Truncate or delete data from the table
        with engine.connect() as connection:
            # Iterate over each table in the database
            for table_name, table in metadata.tables.items():
                # Create a Table object
                t = Table(table_name, metadata, autoload=True, autoload_with=engine)
                # Truncate data from the table (if you want to keep the table structure)
                # connection.execute(t.delete())

                # Alternatively, delete the table itself (if you want to remove the table as well)
                sql_statement = f"DELETE FROM {table_name}"
                result_proxy = connection.execute(text(sql_statement))

            # Commit the changes
            connection.commit()
            # # Get the row count
            # row_count = result_proxy.rowcount
            # print(table_name, row_count)

            # Fetch all rows from the result set
            # rows = result.fetchall()
            engine.dispose()
    def force_initialize(self):
        self._initialize()


import re

from sqlalchemy import Column, Integer, inspect
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm import declarative_base

BaseModelDeclaration = declarative_base()


class BaseModel(BaseModelDeclaration):
    __abstract__ = True
    id = Column(Integer, primary_key=True)
    @classmethod
    def __init_subclass__(cls, **kwargs):
        # Convert class name to snake case and set it as __tablename__
        # cls.__tablename__ = cls.snake_case(cls.__name__).replace("_model", "")
        cls.__tablename__ = cls.snake_case(cls.__name__)
        super().__init_subclass__(**kwargs)

    @staticmethod
    def snake_case(name):
        name = re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()
        return name

    def to_dict(self):
        result = {}

        # Iterate over columns from all parent classes
        for cls in self.__class__.__mro__:
            if hasattr(cls, '__table__'):
                for column in cls.__table__.columns:
                    result[column.name] = str(getattr(self, column.name))

        # Iterate over relationships from all parent classes
        for cls in self.__class__.__mro__:
            if hasattr(cls, '__mapper__'):
                for relationship in class_mapper(cls).relationships:
                    related_obj = getattr(self, relationship.key)
                    if related_obj is not None:
                        if relationship.uselist:
                            result[relationship.key] = [obj.to_dict() for obj in related_obj]
                        else:
                            result[relationship.key] = related_obj.to_dict()

        return result
    # def to_dict(self):
    #     result = {}
    #     for column in self.__table__.columns:
    #         result[column.name] = str(getattr(self, column.name))
    #     # Include polymorphic identity if present
    #     for relationship in class_mapper(self.__class__).relationships:
    #         related_obj = getattr(self, relationship.key)
    #         if related_obj is not None:
    #             if relationship.uselist:
    #                 result[relationship.key] = [obj.to_dict() for obj in related_obj]
    #             else:
    #                 result[relationship.key] = related_obj.to_dict()
    #     return result
    def __repr__(self):
        # return vars(self)
        field_strings = []
        for field, value in self._to_print_dict().items():
            field_strings.append(f'{field}={value}')
        return f"<{self.__class__.__name__} ({', '.join(field_strings)})>"

    def _to_print_dict(self):
        return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs}



from sqlalchemy.types import TypeDecorator

from belenios.custom_types.belenios.ZqType import ZqType


class PrivateKeyType(TypeDecorator):
    impl = ZqType

    def process_bind_param(self, value, dialect):
        return value

    def process_result_value(self, value, dialect):
        return value

    def process_literal_param(self, value, dialect):
        return value



from typing import List

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.CredentialModel import CredentialModel
from belenios.models.belenios.EventModel import EventModel
from belenios.models.belenios.OwnedPartialDecryptionModel import OwnedPartialDecryptionModel
from belenios.models.belenios.PolynomialModel import PolynomialModel
from belenios.models.belenios.ResultModel import ResultModel
from belenios.models.belenios.SaltModel import SaltModel
from belenios.models.belenios.TallyModel import TallyModel
from belenios.models.belenios.TrusteeKindModel import TrusteeKindModel
from belenios.models.belenios.VoterModel import VoterModel, assoc_elections_bundle_and_voters


class ElectionBundleModel(BaseModel):
    election_id = Column(Integer, ForeignKey('election_model.id'), nullable=False)
    election = relationship('ElectionModel', back_populates='election_bundle')
    trustee_kinds: Mapped[List[TrusteeKindModel]] = relationship('TrusteeKindModel', uselist=True,
                                                                     cascade="all, delete-orphan", single_parent=True)
    voters: Mapped[List[VoterModel]] = relationship(secondary=assoc_elections_bundle_and_voters,
                                                    back_populates="election_bundles")
    public_credentials: Mapped[List[CredentialModel]] = relationship('CredentialModel', uselist=True)
    salts: Mapped[List[SaltModel]] = relationship('SaltModel', uselist=True)
    setup_data_id = Column(Integer, ForeignKey('setup_data_model.id'), nullable=False)
    setup_data = relationship('SetupDataModel')
    events: Mapped[List[EventModel]] = relationship('EventModel', uselist=True, cascade="all, delete-orphan",
                                                      single_parent=True)
    tally: Mapped[TallyModel] = relationship('TallyModel')
    owned_partial_decryptions: Mapped[List[OwnedPartialDecryptionModel]] = relationship('OwnedPartialDecryptionModel',
                                                                                          uselist=True,
                                                                                          cascade="all, delete-orphan",
                                                                                          single_parent=True)
    result_id: int = Column(Integer, ForeignKey('result_model.id'), nullable=True)
    result: Mapped[ResultModel] = relationship('ResultModel', uselist=False)
    polynomials: Mapped[List[PolynomialModel]] = relationship('PolynomialModel', uselist=True,
                                                              cascade="all, delete-orphan", single_parent=True)

    def __str__(self):
        trustee_kinds_info = ', '.join(str(trustee_kind) for trustee_kind in self.trustee_kinds)
        voters_info = ', '.join(str(voter) for voter in self.voters)
        public_credentials_info = ', '.join(str(credential) for credential in self.public_credentials)
        salts_info = ', '.join(str(salt) for salt in self.salts)
        events_info = ', '.join(str(event) for event in self.events)
        owned_partial_decryptions_info = ', '.join(
            str(owned_partial_decryption) for owned_partial_decryption in self.owned_partial_decryptions)

        return (
            f"ElectionBundleModel(\n"
            f"    election={self.election},\n"
            f"    trustee_kinds=[{trustee_kinds_info}],\n"
            f"    voters=[{voters_info}],\n"
            f"    public_credentials=[{public_credentials_info}],\n"
            f"    salts=[{salts_info}],\n"
            f"    setup_data={self.setup_data},\n"
            f"    events=[{events_info}],\n"
            f"    tally={self.tally},\n"
            f"    owned_partial_decryptions=[{owned_partial_decryptions_info}],\n"
            f"    result={self.result}\n"
            f")"
        )


from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class SignatureModel(BaseModel):
    ballot_id = Column(Integer, ForeignKey('ballot_model.id'), nullable=False)
    hash = Column(StringType, nullable=False)
    proof_id = Column(Integer, ForeignKey('proof_model.id'), nullable=False)
    proof = relationship('ProofModel')



from typing import List

from sqlalchemy import Column, Integer, ForeignKey, Table
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.AnswerModel import AnswerModel
from belenios.models.belenios.CiphertextModel import CiphertextModel
from belenios.models.belenios.ProofModel import ProofModel
from belenios.utilities.Utility import Utility

assoc_answer_ciphertext= Table(
    "assoc_answer_ciphertext",
    BaseModel.metadata,
    Column("answer_id", Integer, ForeignKey("answer_model.id"), primary_key=True),
    Column("ciphertext_id", Integer, ForeignKey("ciphertext_model.id"), primary_key=True)
)

assoc_answer_i_proof = Table(
    "assoc_answer_i_proof",
    BaseModel.metadata,
    Column("answer_id", Integer, ForeignKey("answer_model.id"), primary_key=True),
    Column("i_proof_id", Integer, ForeignKey("i_proof_model.id"), primary_key=True)
)

class AnswerHModel(AnswerModel):
    id = Column(Integer, ForeignKey('answer_model.id'), primary_key=True)
    choices: Mapped[List[CiphertextModel]] = relationship('CiphertextModel', secondary=assoc_answer_ciphertext)
    individual_proofs = relationship('IProofModel', secondary=assoc_answer_i_proof)
    overall_proof_id: int = Column(Integer, ForeignKey('proof_model.id'), nullable=False)
    overall_proof: Mapped[ProofModel] = relationship('ProofModel', foreign_keys=[overall_proof_id])
    blank_proof1_id: int = Column(Integer, ForeignKey('proof_model.id'), nullable=True)
    blank_proof1: Mapped[ProofModel] = relationship('ProofModel', foreign_keys=[blank_proof1_id])
    blank_proof2_id: int = Column(Integer, ForeignKey('proof_model.id'), nullable=True)
    blank_proof2: Mapped[ProofModel] = relationship('ProofModel', foreign_keys=[blank_proof2_id])
    __mapper_args__ = {
        "polymorphic_identity": "AnswerHModel",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'AnswerHModel'

    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"id={self.id}" + delimiter +
                f"choices={[choice.to_hash_sha256() for choice in self.choices]}" + delimiter +
                f"individual_proofs={[proof.to_hash_sha256() for proof in self.individual_proofs]}" + delimiter +
                f"overall_proof={self.overall_proof.to_hash_sha256()}" + delimiter +
                f"blank_proof1={self.blank_proof1.to_hash_sha256() if self.blank_proof1 else None}" + delimiter +
                f"blank_proof2={self.blank_proof2.to_hash_sha256() if self.blank_proof2 else None}"
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash
    def __str__(self):
        choices_info = ', '.join(str(choice) for choice in self.choices)
        individual_proofs_info = ', '.join(str(individual_proof) for individual_proof in self.individual_proofs)
        return (
            f"AnswerHModel(\n"
            # f"    id={self.id},\n"
            f"    choices=[{choices_info}],\n"
            f"    individual_proofs=[{individual_proofs_info}],\n"
            # f"    overall_proof_id={self.overall_proof_id},\n"
            f"    overall_proof={self.overall_proof},\n"
            # f"    blank_proof1_id={self.blank_proof1_id},\n"
            f"    blank_proof1={self.blank_proof1},\n"
            # f"    blank_proof2_id={self.blank_proof2_id},\n"
            f"    blank_proof2={self.blank_proof2}\n"
            f")"
        )



from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal


from belenios.controllers.default_cmd import DefaultCmd
from belenios.controllers.election_cmd import ElectionCmd
from belenios.controllers.setup_cmd import SetupCmd
from belenios.core.command_line.exc import BeleniosError
from belenios.models.BaseModel import BaseModel
from belenios.utilities.DatabaseInit import DatabaseInit
from belenios.utilities.command_line.AppObject import AppObject


# configuration defaults

def post_setup(app):
    AppObject().addApp(app)


class Belenios(App):
    """Belenios primary application."""

    class Meta:
        label = 'belenios'

        config_files = ['config.yml']

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
            'jinja2',
        ]

        hooks = [
            ('post_setup', post_setup),
        ]
        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        base_controller = DefaultCmd
        # register handlers
        handlers = [
            DefaultCmd,
            SetupCmd,
            ElectionCmd,
        ]


class BeleniosTest(TestApp, Belenios):
    """A sub-class of Belenios that is better suited for testing."""

    class Meta:
        label = 'belenios'


def main():
    with Belenios() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except BeleniosError as e:
            print('BeleniosError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()



from sqlalchemy import Column, Integer, ForeignKey

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class TrusteeKindModel(BaseModel):
    election_bundle_id = Column(Integer, ForeignKey('election_bundle_model.id'), nullable=False)
    trustee_kind_type = Column(StringType, nullable=False)
    __mapper_args__ = {
        "polymorphic_identity": "TrusteeKindModel",
        "polymorphic_on": trustee_kind_type
    }

    def __str__(self):
        return (
            f"TrusteeKindModel(\n"
            # f"    election_bundle_id={self.election_bundle_id},\n"
            f"    trustee_kind_type={self.trustee_kind_type}\n"
            f")"
        )


from belenios.utilities.Utility import Utility


# https://gitlab.inria.fr/belenios/belenios/-/blob/master/src/lib/v1/trustees.ml
class ProofOfKnowledge:
    @staticmethod
    def Hpok(group, public_key, A):
        zkp = ("pok|" + str(group.description) + "|" + str(public_key) + "|" + str(A)).encode()
        res = int.from_bytes(Utility.hash_sha256_as_bytes(zkp), byteorder='big')
        return res

    @staticmethod
    def Hraweg(S0, election, answer, A):
        zkp = ("raweg|" + str(S0) + "|" + str(election.public_key) + "|" + str(answer.choices.alpha) + "|" + str(
            answer.choices.beta) + "|" + str(A)).encode()
        res = int.from_bytes(Utility.hash_sha256_as_bytes(zkp), byteorder='big')
        return res

    @staticmethod
    def Hsignature(H, A):
        zkp = ("sig|" + str(H) + "|" + str(A)).encode()
        res = int.from_bytes(Utility.hash_sha256_as_bytes(zkp), byteorder='big')
        return res

    @staticmethod
    def Hsignmsg(M, A):
        zkp = ("sigmsg|" + str(M) + "|" + str(A)).encode()
        res = int.from_bytes(Utility.hash_sha256_as_bytes(zkp), byteorder='big')
        return res

    @staticmethod
    def Hiprove(S, ciphertext, As, Bs):
        As_Bs_part = ""
        for A, B in zip(As, Bs):
            As_Bs_part += str(A) + ","
            As_Bs_part += str(B) + ","
        As_Bs_part = As_Bs_part[:-1]
        zkp = ("prove|" + str(S) + "|" + str(ciphertext.alpha) + "|" + str(
            ciphertext.beta) + "|" + As_Bs_part).encode()
        res = int.from_bytes(Utility.hash_sha256_as_bytes(zkp), byteorder='big')
        return res



from typing import List

from sqlalchemy import Column, Integer, ForeignKey, Table
from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import event
from belenios.custom_types.belenios.ArrayOfIsType import ArrayOfIsType
from belenios.custom_types.belenios.IType import IType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.SignedMsgModel import SignedMsgModel
from belenios.models.belenios.TrusteeKindModel import TrusteeKindModel
from belenios.models.belenios.TrusteePublicKeyModel import TrusteePublicKeyModel
from belenios.utilities.command_line.DbSession import DbSession

assoc_pedersen_trustee_certs = Table(
    "assoc_pedersen_trustee_certs",
    BaseModel.metadata,
    Column("pedersen_trustee_id", Integer, ForeignKey("pedersen_trustee_model.id"), primary_key=True),
    Column("cert_id", Integer, ForeignKey("signed_msg_model.id"), primary_key=True)
)
assoc_pedersen_trustee_coefexps = Table(
    "assoc_pedersen_trustee_coefexps",
    BaseModel.metadata,
    Column("pedersen_trustee_id", Integer, ForeignKey("pedersen_trustee_model.id"), primary_key=True),
    Column("coefexp_id", Integer, ForeignKey("signed_msg_model.id"), primary_key=True)
)
assoc_pedersen_trustee_verification_keys = Table(
    "assoc_pedersen_trustee_verification_keys",
    BaseModel.metadata,
    Column("pedersen_trustee_id", Integer, ForeignKey("pedersen_trustee_model.id"), primary_key=True),
    Column("verification_key_id", Integer, ForeignKey("trustee_public_key_model.id"), primary_key=True)
)


class PedersenTrusteeModel(TrusteeKindModel):
    id = Column(Integer, ForeignKey('trustee_kind_model.id'), primary_key=True)
    threshold = Column(IType, nullable=False)
    trustee_ids: Mapped[List[IType]] = Column(ArrayOfIsType, nullable=False)
    certs: Mapped[List[SignedMsgModel]] = relationship('SignedMsgModel', uselist=True,
                                                       secondary=assoc_pedersen_trustee_certs)
    coefexps: Mapped[List[SignedMsgModel]] = relationship('SignedMsgModel', uselist=True,
                                                          secondary=assoc_pedersen_trustee_coefexps)
    verification_keys: Mapped[List[TrusteePublicKeyModel]] = relationship('TrusteePublicKeyModel', uselist=True,
                                                                          secondary=assoc_pedersen_trustee_verification_keys)

    __mapper_args__ = {
        "polymorphic_identity": "PedersenTrusteeModel"
    }

    def __init__(self, *args, **kwargs):
        trustee_ids = kwargs.pop('trustee_ids', None)
        super().__init__(*args, **kwargs)
        self.trustee_kind_type = 'PedersenTrusteeModel'
        self.trustee_ids = trustee_ids if trustee_ids is not None else []

    def append_trustee_id(self, trustee_id):
        if trustee_id not in self.trustee_ids:
            a = [i for i in self.trustee_ids]
            a.append(trustee_id)
            self.trustee_ids = a



from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel


class VoterCredentialModel(BaseModel):
    election_id = Column(Integer, ForeignKey('election_model.id'), nullable=False)
    election: Mapped['ElectionModel'] = relationship("ElectionModel")
    voter_id = Column(Integer, ForeignKey('voter_model.id'), nullable=False)
    voter: Mapped['VoterModel'] = relationship("VoterModel")
    credential_id = Column(Integer, ForeignKey('credential_model.id'), nullable=False)
    credential: Mapped['CredentialModel'] = relationship("CredentialModel")



from sqlalchemy import Column

from belenios.custom_types.belenios.GType import GType
from belenios.models.BaseModel import BaseModel
from belenios.utilities.Utility import Utility


class CiphertextModel(BaseModel):
    alpha = Column(GType, nullable=False)
    beta = Column(GType, nullable=False)

    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"alpha={self.alpha}" + delimiter +
                f"beta={self.beta}" + delimiter
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash

    def __str__(self):
        return (
            f"CiphertextModel(\n"
            f"    alpha={self.alpha},\n"
            f"    beta={self.beta}\n"
            f")"
        )


from dataclasses import dataclass
from typing import List, Tuple


@dataclass
class KeysPair:
    public_key: int
    private_key: int


@dataclass
class Group:
    description: str
    p: int
    g: int
    q: int
@dataclass
class EncryptedText:
    chunk: List[Tuple[int, int]] # (u, v)
@dataclass
class EncryptedTextAndK:
    chunk: List[Tuple[int, int, int]] # ( u, v, k )


@dataclass
class PrivateCredential:
    index: int
    private_credential: str
    email: str



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.GType import GType
from belenios.custom_types.belenios.NType import NType
from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.EmbeddingModel import EmbeddingModel
from belenios.utilities.Utility import Utility


class GroupModel(BaseModel):
    description = Column(StringType, nullable=False)
    g = Column(GType, nullable=False)
    p = Column(NType, nullable=False)
    q = Column(NType, nullable=False)
    election_id: int = Column(Integer, ForeignKey('election_model.id'), nullable=False)
    election: Mapped['ElectionModel'] = relationship('ElectionModel', back_populates='group')
    embedding: Mapped["EmbeddingModel"] = relationship('EmbeddingModel', back_populates='group',
                                                       uselist=False, cascade="all, delete-orphan",
                                                       single_parent=True)
    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"description={self.description}" + delimiter +
                f"g={self.g}" + delimiter +
                f"p={self.p}" + delimiter +
                f"q={self.q}"
            # Add embedding if needed
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash

    def __str__(self):
        return (
            f"GroupModel(\n"
            f"    description={self.description},\n"
            f"    g={self.g},\n"
            f"    p={self.p},\n"
            f"    q={self.q},\n"
            # f"    election_id={self.election_id},\n"
            # f"    election={self.election},\n"
            f"    embedding={self.embedding}\n"
            f")"
        )


from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.ProofModel import ProofModel
from belenios.models.belenios.SignedObjectModel import SignedObjectModel


class SignedMsgModel(BaseModel):
    signed_object_id = Column(Integer, ForeignKey('signed_object_model.id'), nullable=False)
    signed_object: Mapped[SignedObjectModel] = relationship('SignedObjectModel', uselist=False)
    signature_id = Column(Integer, ForeignKey('proof_model.id'), nullable=False)
    signature: Mapped[ProofModel] = relationship('ProofModel', uselist=False)

    # def to_json(self):
    #     return {
    #         'signed_object': self.signed_object.to_json(),
    #         'signature': self.signature.to_json(),
    #     }

    @staticmethod
    def load_from_json(dict):
        return SignedMsgModel(signed_object=SignedObjectModel.load_from_json(dict['signed_object']),
                              signature=ProofModel.load_from_json(dict['signature']))



from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey, Table
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel

assoc_encrypted_tally_row_and_ciphertext = Table(
    "assoc_encrypted_tally_row_and_ciphertext",
    BaseModel.metadata,
    Column("encrypted_tally_row_model_id", ForeignKey("encrypted_tally_row_model.id"), primary_key=True),
    Column("ciphertext_model_id", ForeignKey("ciphertext_model.id"), primary_key=True),
)


class EncryptedTallyRowModel(BaseModel):
    tally_id = Column(Integer, ForeignKey('tally_model.id'))
    tally: Mapped['TallyModel'] = relationship('TallyModel')
    ciphertexts = relationship("CiphertextModel", secondary=assoc_encrypted_tally_row_and_ciphertext)



from sqlalchemy.types import TypeDecorator, VARCHAR


class GType(TypeDecorator):
    impl = VARCHAR
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = str(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = int(value)
        return value

    def process_literal_param(self, value, dialect):
        return value


import json

from cement import Controller, ex
from cement.utils import shell

from belenios.core.belenios.CredentialAuthorityCore import CredentialAuthorityCore
from belenios.core.belenios.ServerAdministratorCore import ServerAdministratorCore
from belenios.core.belenios.ServerCore import ServerCore
from belenios.core.belenios.TrusteeCore import TrusteeCore
from belenios.models.belenios.ProofModel import ProofModel
from belenios.models.belenios.QuestionHModel import QuestionHModel
from belenios.models.belenios.QuestionNhModel import QuestionNhModel
from belenios.models.belenios.TrusteePublicKeyModel import TrusteePublicKeyModel
from belenios.models.belenios.VInputModel import VInputModel


class SetupCmd(Controller):
    class Meta:
        label = 'setup'
        stacked_on = 'base'
        stacked_type = 'nested'
        # base_controller = True

        # text displayed at the top of --help output
        description = 'Belenios command-line tool to create and setup an election.'

        # text displayed at the bottom of --help output
        epilog = 'Usage: belenios setup'

        arguments = [
            # (['-start', '--start'],
            #  {'action': 'start',
            #   '': }),
        ]

    # @ex(
    #     help='login',
    #
    #     # sub-command level arguments.
    #     arguments=[
    #         (['-u', '--username'],
    #          {'help': 'The username to login', 'action': 'store', 'dest': 'username', 'required': True}),
    #         (['-p', '--password'],
    #          {'help': 'The password to login', 'action': 'store', 'dest': 'password', 'required': True}),
    #     ],
    # )
    # def login(self):
    #     print("login :", self.app.pargs.username, self.app.pargs.password)
    #     # Retrieve username and password from command-line arguments
    #     username = self.app.pargs.username
    #     password = self.app.pargs.password
    #
    #     # Store the session information
    #     print(UserSession().username)
    #
    #     print("Logged in successfully!")

    @ex(
        help='Start a new election creation',

        # sub-command level arguments.
        arguments=[
            (['-n', '--name'],
             {'help': 'The name of the election', 'action': 'store', 'dest': 'name'}),
            (['-d', '--description'],
             {'help': 'The description of the election', 'action': 'store', 'dest': 'description'}),
            (['-g', '--group'], {'help': 'The group of the election', 'action': 'store', 'dest': 'group'}),
            (['-a', '--administrator'],
             {'help': 'The email of the administrator of the election', 'action': 'store', 'dest': 'administrator'}),
            (['-c', '--credential_authority'],
             {'help': 'The email of the credential authority of the election', 'action': 'store',
              'dest': 'credential_authority'}),
        ],
    )
    def start(self):
        if not self.app.pargs.name:
            self.app.pargs.name = shell.Prompt("Enter the name of the election: ").input

        if not self.app.pargs.description:
            self.app.pargs.description = shell.Prompt("Enter the description of the election: ").input

        if not self.app.pargs.group:
            self.app.pargs.group = shell.Prompt("Enter the group of the election: ").input
        self.app.pargs.group = self.app.pargs.group.upper()

        if not self.app.pargs.administrator:
            self.app.pargs.administrator = shell.Prompt("Enter the email of the administrator: ").input

        if not self.app.pargs.credential_authority:
            self.app.pargs.credential_authority = shell.Prompt("Enter the email of the credential authority: ").input

        # Now you can proceed with the creation of the election using the provided inputs
        print("Starting election creation with the following parameters:")
        print("Name: %s" % self.app.pargs.name)
        print("Description: %s" % self.app.pargs.description)
        print("Group: %s" % self.app.pargs.group)
        print("Administrator: %s" % self.app.pargs.administrator)
        print("Credential Authority: %s" % self.app.pargs.credential_authority)

        election_bundle = ServerAdministratorCore.start_new_election(self.app.pargs.name, self.app.pargs.description,
                                                                     self.app.pargs.administrator,
                                                                     self.app.pargs.credential_authority,
                                                                     self.app.pargs.group)
        if election_bundle is None:
            self.print("Election not created")
            return False
        ServerCore().make_election_unique(election_bundle.election)

        print("New election created with uuid: %s" % election_bundle.election.uuid)

        return True

    @ex(
        help='Add a new voter to the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'],
             {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-e', '--email'],
             {'help': 'The email of the voter', 'action': 'store', 'dest': 'email'}),
            (['-w', '--weight'],
             {'help': 'The weight of the voter', 'action': 'store', 'dest': 'weight'}),
        ],
    )
    def add_voter(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.email:
            self.app.pargs.email = shell.Prompt("Enter the email of the new voter: ").input

        if not self.app.pargs.weight:
            self.app.pargs.weight = shell.Prompt("Enter the weight of the new voter: ").input

        print("Starting to add a new voter with the following parameters:")
        print("Uuid: %s" % self.app.pargs.uuid)
        print("Email: %s" % self.app.pargs.email)
        print("Weight: %s" % self.app.pargs.weight)

        voter = ServerAdministratorCore.add_voter(self.app.pargs.uuid, self.app.pargs.email,
                                                  self.app.pargs.weight, )
        if voter is None:
            self.print("Voter not created")
            return False

        print("New voter created.")
        return True

    @ex(
        help='Generate voters credentials and salts for the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-fprc', '--file-private-credentials'],
             {'help': 'The file where to output private credentials of the election', 'action': 'store',
              'dest': 'file_private_credentials'}),
            (['-fpuc', '--file-public-credentials'],
             {'help': 'The file where to output public credentials of the election', 'action': 'store',
              'dest': 'file_public_credentials'}),
            (['-fs', '--file--salts'],
             {'help': 'The file where to output salts of the election', 'action': 'store', 'dest': 'file_salts'}),
        ],
    )
    def generate_credentials(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.file_private_credentials:
            self.app.pargs.file_private_credentials = shell.Prompt(
                "Enter the file where to output private credentials of the election: ").input

        if not self.app.pargs.file_public_credentials:
            self.app.pargs.file_public_credentials = shell.Prompt(
                "Enter the file where to output public credentials of the election: ").input

        if not self.app.pargs.file_salts:
            self.app.pargs.file_salts = shell.Prompt("Enter the file where to output salts of the election: ").input

        print("Starting generation parameters:")
        print("uuid: %s" % self.app.pargs.uuid)

        # sending election.uuid to C
        private_credentials, public_credentials, salts = CredentialAuthorityCore().generate_voters_credentials_and_salts(
            self.app.pargs.uuid)

        if private_credentials is None or public_credentials is None or salts is None or len(salts) == 0 or len(
                private_credentials) != len(public_credentials) or len(public_credentials) != len(salts):
            print("Credentials and salts were not generated")
            return False

        if ServerCore().check_voters_credentials_and_salts(self.app.pargs.uuid, public_credentials, salts):
            print("The voting server checked email, weight, public credentials and salts.")
        else:
            print("The voting server detected an error in public data")
            return False

        with open(self.app.pargs.file_private_credentials, 'w') as file:
            json.dump(private_credentials, file)

        with open(self.app.pargs.file_public_credentials, 'w') as file:
            json.dump(public_credentials, file)

        with open(self.app.pargs.file_salts, 'w') as file:
            json.dump(salts, file)

        print("Credentials and salts generated for %s voters." % len(private_credentials))
        return True

    @ex(
        help='Generate a trustees key for the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-fprk', '--file-private-key'],
             {'help': 'The file where to output private key of the trustee', 'action': 'store',
              'dest': 'file_private_key'}),
            (['-fpuk', '--file-public-key'],
             {'help': 'The file where to output public key of the trustee', 'action': 'store',
              'dest': 'file_public_key'}),
        ],
    )
    def generate_trustee_key(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.file_private_key:
            self.app.pargs.file_private_key = shell.Prompt(
                "Enter file where to output private key of the trustee: ").input

        if not self.app.pargs.file_public_key:
            self.app.pargs.file_public_key = shell.Prompt(
                "Enter file where to output public key of the trustee: ").input

        print("Starting generation parameters:")
        print("uuid: %s" % self.app.pargs.uuid)

        # trustees = trustee_kind∗
        # trustee_kind = ["Single", trustee_public_key] | ["Pedersen", threshold_parameters]
        # [["Single", . . . ], ["Single", . . . ], ["Single", . . . ]]
        # 8. S defines the shape of the trustees structure that will be used in the election depending
        # on A’s instructions;
        # TODO: use A’s instructions;

        # 9. S and T1, . . . , Tm run key establishment protocols (see 3.1.1) as needed to fill in the trustees
        # structure;

        # "Single" protocol :

        # 1. T generates a trustee_public_key γ and sends it to S
        trustee_public_key, keys_pair = TrusteeCore().generates_single_trustee_public_key(self.app.pargs.uuid)
        if trustee_public_key is None or keys_pair is None:
            print("The key was not created.")

        with open(self.app.pargs.file_private_key, 'w') as file:
            json.dump(keys_pair.private_key, file)

        with open(self.app.pargs.file_public_key, 'w') as file:
            json.dump(trustee_public_key.to_json(), file)
        print("Trustee key generated.")
        return True

    @ex(
        help='Generate a threshold trustees key for the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-fprk', '--file-private-key'],
             {'help': 'The file where to output private key of the trustee', 'action': 'store',
              'dest': 'file_private_key'}),
            (['-fc', '--file-cert'],
             {'help': 'The file where to output cert of the trustee', 'action': 'store',
              'dest': 'file_cert'}),
            (['-s', '--step'],
             {'help': 'The step of the Perdersen trustee key generation', 'action': 'store',
              'dest': 'step'}),
            (['-tid', '--trustee-id'],
             {'help': 'The trustee id for the election', 'action': 'store', 'dest': 'trustee_id'}),
            (['-t', '--threshold'],
             {'help': 'The threshold need for trustee key decryption', 'action': 'store', 'dest': 'threshold'}),
            (['-vi', '--vinput'],
             {'help': 'The file where to input the verification data for trustee key decryption', 'action': 'store',
              'dest': 'file_vinput'}),
            (['-vo', '--voutput'],
             {'help': 'The file where to output the verification data for trustee key decryption', 'action': 'store',
              'dest': 'file_voutput'}),
            (['-dk', '--decryption-key'],
             {'help': 'The file where to output the decryption key for trustee key decryption', 'action': 'store',
              'dest': 'file_decryption_key'}),
        ],
    )
    def generate_trustee_key_threshold(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.file_private_key:
            self.app.pargs.file_private_key = shell.Prompt(
                "Enter file where to output private key of the trustee: ").input

        if not self.app.pargs.step:
            self.app.pargs.step = shell.Prompt("Enter step of the Perdersen trustee key generation: ").input
        self.app.pargs.step = int(self.app.pargs.step)

        if not self.app.pargs.trustee_id:
            self.app.pargs.trustee_id = shell.Prompt("Enter the trustee id for the election: ").input
        self.app.pargs.trustee_id = int(self.app.pargs.trustee_id)

        print("Starting generation parameters:")
        print("uuid: %s" % self.app.pargs.uuid)
        print("step: %s" % self.app.pargs.step)

        # trustees = trustee_kind∗
        # trustee_kind = ["Single", trustee_public_key] | ["Pedersen", threshold_parameters]
        # [["Single", . . . ], ["Single", . . . ], ["Single", . . . ]]
        # 8. S defines the shape of the trustees structure that will be used in the election depending
        # on A’s instructions;
        # TODO: use A’s instructions;

        # 9. S and T1, . . . , Tm run key establishment protocols (see 3.1.1) as needed to fill in the trustees
        # structure;

        # "Perdersen" protocol :

        if self.app.pargs.step == 1:  # and 2 actually

            if not self.app.pargs.file_cert:
                self.app.pargs.file_cert = shell.Prompt(
                    "Enter file where to output cert of the trustee: ").input

            # (a) Tz generates a cert γz and sends it to S
            signed_msg, seed = TrusteeCore().generates_perdersen_trustee_keys(self.app.pargs.uuid)
            if signed_msg is None or seed is None:
                print("The cert was not created.")
            print("The cert has been created.")

            # ttkeygen --certs certs.jsons --step 2
            pedersen_trustee = ServerCore.import_cert_pedersen_trustee(self.app.pargs.uuid, signed_msg,
                                                                     self.app.pargs.trustee_id)
            if pedersen_trustee is None:
                print("The cert was not imported.")
                return False
            print("The cert has been imported.")

            with open(self.app.pargs.file_private_key, 'w') as file:
                json.dump(seed, file)

            with open(self.app.pargs.file_cert, 'w') as file:
                json.dump(signed_msg.to_dict(), file)
        elif self.app.pargs.step == 3:
            if not self.app.pargs.threshold:
                self.app.pargs.threshold = shell.Prompt("Enter the threshold need for trustee key decryption: ").input
            self.app.pargs.threshold = int(self.app.pargs.threshold)
            try:
                with open(self.app.pargs.file_private_key, 'r') as file:
                    seed = json.load(file)
            except:
                print("Invalid private key file")
                return False

            polynomial = TrusteeCore.generates_perdersen_trustee_polynomial(self.app.pargs.uuid,
                                                                            self.app.pargs.threshold, seed,
                                                                            self.app.pargs.trustee_id)
            if polynomial is None:
                print("The polynomial was not created.")
                return False
            print("The polynomial has been created.")
        elif self.app.pargs.step == 4:
            if not self.app.pargs.file_vinput:
                self.app.pargs.file_vinput = shell.Prompt("Enter file where to input the verification data: ").input

            # if not self.app.pargs.threshold:
            #     self.app.pargs.threshold = shell.Prompt("Enter the threshold need for trustee key decryption: ").input
            # self.app.pargs.threshold = int(self.app.pargs.threshold)
            # if not self.app.pargs.file_voutput:
            #     self.app.pargs.file_voutput = shell.Prompt("Enter file where to output the verification data: ").input
            try:
                with open(self.app.pargs.file_private_key, 'r') as file:
                    seed = json.load(file)
            except:
                print("Invalid private key file")
                return False
            vinput = TrusteeCore.generates_perdersen_trustee_vinput(self.app.pargs.uuid, seed,
                                                                    self.app.pargs.trustee_id)

            if vinput is None:
                print("The vinput was not created.")
                return False
            print("The vinput has been created.")
            check = TrusteeCore.check_perdersen_trustee_vinput(self.app.pargs.uuid, seed, self.app.pargs.trustee_id,
                                                               vinput)

            if check != True:
                print("The vinput was not correctly generated.")
                return False

            with open(self.app.pargs.file_vinput, 'w') as file:
                json.dump(vinput.to_dict(), file)
        elif self.app.pargs.step == 5:
            if not self.app.pargs.file_vinput:
                self.app.pargs.file_vinput = shell.Prompt("Enter file where to input the verification data: ").input

            if not self.app.pargs.file_voutput:
                self.app.pargs.file_voutput = shell.Prompt("Enter file where to output the verification data: ").input

            if not self.app.pargs.file_decryption_key:
                self.app.pargs.file_decryption_key = shell.Prompt(
                    "Enter file where to output the decryption key: ").input
            try:
                with open(self.app.pargs.file_private_key, 'r') as file:
                    seed = json.load(file)
            except:
                print("Invalid private key file")
                return False

            try:
                with open(self.app.pargs.file_vinput, 'r') as file:
                    vinput_json = json.load(file)

            except Exception as e:
                print("Invalid private key file")
                print(e)
                return False
            v_input = VInputModel.load_from_json(vinput_json)
            voutput, private_key = TrusteeCore.generates_perdersen_trustee_voutput(self.app.pargs.uuid, seed,
                                                                      self.app.pargs.trustee_id, v_input)

            if voutput is None:
                print("The voutput was not created.")
                return False
            print("The voutput has been created.")

            check = ServerAdministratorCore.check_perdersen_trustee_voutput(self.app.pargs.uuid,
                                                                            self.app.pargs.trustee_id, voutput)

            if check != True:
                print("The voutput was not correctly generated.")
                return False

            pedersen_trustee = ServerCore.import_coefexps_verification_keys_pedersen_trustee(self.app.pargs.uuid,
                                                                                             voutput)
            if pedersen_trustee is None:
                print("The verification_key was not imported.")
                return False
            print("The verification_key has been imported.")

            with open(self.app.pargs.file_voutput, 'w') as file:
                json.dump(voutput.to_dict(), file)
            with open(self.app.pargs.file_decryption_key, 'w') as file:
                json.dump(private_key, file)
        #     TODO: admin check public_key
        # TODO: check if step 6 is already done
        else:
            print("step invalid")
            return False
        return True

    @ex(
        help='Send a trustees key to the server for the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-fpuk', '--file-public-key'],
             {'help': 'The file where to output public key of the trustee', 'action': 'store',
              'dest': 'file_public_key'}),
            (['-tid', '--trustee-id'],
             {'help': 'The trustee id for the election', 'action': 'store', 'dest': 'trustee_id'}),
        ],
    )
    def send_trustee_key(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.file_public_key:
            self.app.pargs.file_public_key = shell.Prompt(
                "Enter file where to output public key of the trustee: ").input

        if not self.app.pargs.trustee_id:
            self.app.pargs.trustee_id = shell.Prompt("Enter the id of the trustee for the election: ").input

        print("Starting generation parameters:")
        print("uuid: %s" % self.app.pargs.uuid)

        try:
            with open(self.app.pargs.file_public_key, 'r') as file:
                data = json.load(file)
            trustee_public_key = TrusteePublicKeyModel()
            trustee_public_key.public_key = data['public_key']
            proof = ProofModel()
            proof.challenge = data["pok"]['challenge']
            proof.response = data["pok"]['response']
            trustee_public_key.pok = proof
        except:
            print("Invalid file")
            return False
        single_trustee = ServerCore.import_public_key_single_trustee(self.app.pargs.uuid, trustee_public_key,
                                                                     self.app.pargs.trustee_id)
        if single_trustee is None:
            print("The key was not imported.")
            return False
        print("The key has been imported.")
        return True

    @ex(
        help='Make the election structure and calculate hash to make it immutable',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-t', '--template'],
             {'help': 'The template of the questions of the election', 'action': 'store', 'dest': 'template'}),
        ],
    )
    def make_election(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.template:
            self.app.pargs.template = shell.Prompt("Enter the template of the questions of the election: ").input

        print("Starting generation parameters:")
        print("uuid: %s" % self.app.pargs.uuid)
        print("template: %s" % self.app.pargs.template)

        election_bundle = ServerCore.compute_election_public_key(self.app.pargs.uuid)
        if election_bundle is None:
            print("The election was not created.")
        questions = []
        try:
            with open(self.app.pargs.template, 'r') as file:
                data = json.load(file)
            for i, question_data in enumerate(data):
                print("Adding question:", question_data)
                if question_data.get('type', 'QuestionHModel') == 'QuestionNhModel':
                    question = QuestionNhModel()
                    question.answers = question_data["answers"]
                    question.question = question_data["question"]
                else:
                    question = QuestionHModel()
                    question.answers = question_data["answers"]
                    question.min = question_data.get("min", 0)
                    question.max = question_data.get("max", 0)
                    question.blank = question_data.get("blank", False)
                    question.question = question_data.get("question", "Question?")
                questions.append(question)
        except:
            print("Invalid file")
            return False

        questions = ServerCore.add_questions(self.app.pargs.uuid, questions)
        if questions is None:
            print("The election was not created.")

        election_bundle = ServerCore.make_election_immutable(self.app.pargs.uuid)
        if election_bundle is None:
            print("The election was not created.")

        print("The election has been created.")
        # TODO: handle adding it to event
        return True



from typing import List

from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.ChannelMsgModel import ChannelMsgModel
from belenios.models.belenios.SignedMsgModel import SignedMsgModel


from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey, Table
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.ChannelMsgModel import ChannelMsgModel
from belenios.models.belenios.SignedMsgModel import SignedMsgModel
from belenios.models.belenios.TrusteePublicKeyModel import TrusteePublicKeyModel


class VOutputModel(BaseModel):
    private_key_id = Column(Integer, ForeignKey('channel_msg_model.id'), nullable=False)
    private_key: Mapped[ChannelMsgModel] = relationship('ChannelMsgModel', uselist=False)
    trustee_public_key_id: int = Column(Integer, ForeignKey('trustee_public_key_model.id'), nullable=False)
    trustee_public_key: Mapped[TrusteePublicKeyModel] = relationship('TrusteePublicKeyModel', uselist=False)


import ast

import base58
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Random import get_random_bytes
from Crypto.Random import random
from Crypto.Util.number import inverse
from sqlalchemy.orm import joinedload

from belenios.dataclass.dataclass import KeysPair
from belenios.models.belenios.ChannelMsgModel import ChannelMsgModel
from belenios.models.belenios.EncryptedDataModel import EncryptedDataModel
from belenios.models.belenios.SignedStringModel import SignedStringModel


class Utility:
    # Generating large random numbers in [1, p]
    @staticmethod
    def generate_number_in_q(group):
        return random.randint(1, group.q)

    @staticmethod
    def gen_key_el_gamal(group):
        private_key = random.randint(1, group.q)
        public_key = pow(group.g, private_key, group.p)
        return KeysPair(public_key, private_key)
    @staticmethod
    def generate_base58_string(length):
        # Determine the minimum length required for 14 characters
        min_length = max(length, 14)

        # Generate random bytes
        random_bytes = get_random_bytes(min_length)

        # Encode random bytes to Base58
        base58_string = base58.b58encode(random_bytes)

        # If the length of the resulting Base58 string is less than 14, generate more random bytes and concatenate
        while len(base58_string) < 14:
            additional_bytes = get_random_bytes(min_length)
            base58_string += base58.b58encode(additional_bytes)

        # Truncate the string to the desired length
        return base58_string[:length].decode()
    @staticmethod
    def hash_sha256_as_bytes(data):
        """
        Compute the SHA-256 hash of the input data.

        Args:
            data (bytes): The input data to hash.

        Returns:
            bytes: The SHA-256 hash value.
        """
        hash_object = SHA256.new()
        hash_object.update(data)
        return hash_object.digest()

    @staticmethod
    def hash_sha256_as_hex(data):
        """
        Compute the SHA-256 hash of the input data.

        Args:
            data (bytes): The input data to hash.

        Returns:
            str(hex): The SHA-256 hash value.
        """
        hash_object = SHA256.new()
        hash_object.update(data)
        return hash_object.hexdigest()

    @staticmethod
    def derive_secret_exponent(prefix, salt_model, election):
        # Concatenate the uuid of the election and salt
        salt = election.uuid + salt_model.salt_value
        # Derive the secret exponent x using PBKDF2
        iterations = 100000
        key_length = 64  # 512 bits = 64 bytes
        x = PBKDF2(prefix.encode(), salt.encode(), dkLen=key_length, count=iterations, hmac_hash_module=SHA256)

        # Convert the derived key to an integer in big endian
        x_int = int.from_bytes(x, byteorder='big')

        # Reduce modulo p to form x
        x_mod_p = x_int % election.group.q

        return x_mod_p
    @staticmethod
    def send_mail(app, sender_email, receiver_email, message):
        print('Sending Email :\n\n\t Sender: %s\n\t Receiver: %s\n\t Message: %s\n' % (
        sender_email, receiver_email, message))

    @staticmethod
    def get_election_bundle_from_uuid(election_uuid):
        from belenios.models.belenios.ElectionBundleModel import ElectionBundleModel
        from belenios.models.belenios.ElectionModel import ElectionModel
        from belenios.utilities.command_line.DbSession import DbSession
        return DbSession().session.query(ElectionBundleModel) \
            .join(ElectionModel) \
            .filter(ElectionModel.uuid == election_uuid) \
            .options(joinedload(ElectionBundleModel.election)) \
            .first()

    @staticmethod
    def get_prefix_index_from_credential(private_credential):
        if len(private_credential) != 22:
            return None, None
        prefix = private_credential[:17].replace("-", "")  # XXX-XXXX-XXX-XXXX-NNNN
        index = private_credential[17:].replace("-", "")  # XXX-XXXX-XXX-XXXX-NNNN
        return prefix, int(index)

    @staticmethod
    def put_new_event(election_bundle):
        from belenios.models.belenios.EventModel import EventModel
        event = EventModel()
        # TODO: get last event
        event.height = 0

        election_bundle.events.append(event)
        return event

    @staticmethod
    def generate_signed_msg(signed_object, sk, election):
        from belenios.utilities.ProofOfKnowledge import ProofOfKnowledge
        from belenios.models.belenios.ProofModel import ProofModel
        from belenios.models.belenios.SignedMsgModel import SignedMsgModel
        signed_msg = SignedMsgModel()
        signed_msg.signed_object = signed_object

        # we generate A = g^w using a key pair, where private_key = A and public_key = w
        tmp = Utility.gen_key_el_gamal(election.group)
        w = tmp.private_key
        A = tmp.public_key

        challenge = ProofOfKnowledge.Hsignmsg(str(signed_object), A) % election.group.q
        response = (w - challenge * sk)
        signed_msg.signature = ProofModel()
        signed_msg.signature.challenge = challenge
        signed_msg.signature.response = response
        return signed_msg

    @staticmethod
    def check_signed_msg(signed_msg, vk, election):
        from belenios.utilities.ProofOfKnowledge import ProofOfKnowledge
        left = pow(election.group.g, signed_msg.signature.response, election.group.p)
        right = pow(vk, signed_msg.signature.challenge, election.group.p)
        A = (left * right) % election.group.p
        challenge = ProofOfKnowledge.Hsignmsg(str(signed_msg.signed_object), A) % election.group.q
        return signed_msg.signature.challenge == challenge

    @staticmethod
    def aes_ccm_encrypt(plaintext, key, iv):
        cipher = AES.new(key, AES.MODE_CCM, nonce=iv)
        ciphertext, tag = cipher.encrypt_and_digest(plaintext)
        return ciphertext, tag

    @staticmethod
    def aes_ccm_decrypt(ciphertext, tag, key, iv):
        cipher = AES.new(key, AES.MODE_CCM, nonce=iv)
        plaintext = cipher.decrypt_and_verify(ciphertext, tag)
        return plaintext

    @staticmethod
    def generate_encrypted_msg(plaintext, ek, election):
        from belenios.models.belenios.EncryptedMsgModel import EncryptedMsgModel
        encrypted_msg = EncryptedMsgModel()
        # TODO: use ElGamal() if possible
        # we generate g^r using a key pair, where private_key = r and public_key = alpha
        tmp = Utility.gen_key_el_gamal(election.group)
        r = tmp.private_key
        encrypted_msg.alpha = tmp.public_key

        # we generate g^s using a key pair, where private_key = s and public_key = teta
        tmp = Utility.gen_key_el_gamal(election.group)
        s = tmp.private_key
        teta = tmp.public_key

        encrypted_msg.beta = pow(ek, r, election.group.p) * teta

        key = SHA256.new(("key|" + str(teta)).encode()).digest()
        # print("teta", teta)
        # print("ek", ek)
        # print("key", key)
        iv = SHA256.new(("iv|" + str(encrypted_msg.alpha)).encode()).digest() [:13]
        ciphertext, tag = Utility.aes_ccm_encrypt(plaintext, key, iv)
        # print("fir", Utility.aes_ccm_decrypt(ciphertext,
        #                         tag, key, iv))
        # print("Ciphertext:", ciphertext.hex(), ciphertext == bytes.fromhex(ciphertext.hex()))
        # print("Tag:", tag.hex())

        encrypted_msg.data = EncryptedDataModel()
        encrypted_msg.data.ciphertext = ciphertext.hex()
        encrypted_msg.data.tag = tag.hex()
        encrypted_msg.data.iv = iv.hex()

        return encrypted_msg

    @staticmethod
    def decrypt_encrypted_msg(encrypted_msg, dk, election):
        teta = (encrypted_msg.beta * inverse(pow(encrypted_msg.alpha, dk, election.group.p),
                                             election.group.p)) % election.group.p
        # print("teta", teta)
        key = SHA256.new(("key|" + str(teta)).encode()).digest()
        # print("dk", dk)
        # print("key", key)
        iv = SHA256.new(("iv|" + str(encrypted_msg.alpha)).encode()).digest()[:13]
        if bytes.fromhex(encrypted_msg.data.iv) != iv:
            print("Invalid IV")
            return None
        # print("Ciphertext:", encrypted_msg.data.ciphertext)

        plaintext = Utility.aes_ccm_decrypt(bytes.fromhex(encrypted_msg.data.ciphertext),
                                            bytes.fromhex(encrypted_msg.data.tag), key, iv)
        # print("plaintext:", plaintext)
        return plaintext

    @staticmethod
    def generate_channel_msg(plaintext, sk, ek, election):
        channel_msg = ChannelMsgModel()
        channel_msg.recipient = ek
        signed_string = SignedStringModel()
        signed_string.value = plaintext
        sign = str(Utility.generate_signed_msg(signed_string, sk, election).to_dict()).encode()
        channel_msg.message = Utility.generate_encrypted_msg(sign, ek, election)
        return channel_msg

    @staticmethod
    def decrypt_channel_msg(channel_msg, vk, dk, election):
        from belenios.models.belenios.SignedMsgModel import SignedMsgModel

        decrypted = Utility.decrypt_encrypted_msg(channel_msg.message, dk, election)
        # print("decrypted", decrypted)
        # Convert byte string to string
        byte_string_decoded = decrypted.decode()

        # Use ast.literal_eval() to convert string to dictionary
        result_dict = ast.literal_eval(byte_string_decoded)
        signed_msg = SignedMsgModel.load_from_json(result_dict)
        # print("decrypted fe",result_dict)
        # print("decrypted", signed_msg)
        if not Utility.check_signed_msg(signed_msg, vk, election):
            print("invalid signature")
            return None
        return signed_msg
        #
        # channel_msg = ChannelMsgModel()
        # channel_msg.recipient = ek
        # signed_string = SignedStringModel()
        # signed_string.value = plaintext
        # sign = str(Utility.generate_signed_msg(signed_string, sk, election).to_json()).encode()
        # channel_msg.message = Utility.generate_encrypted_msg(sign, ek, election)
        # return channel_msg



    @staticmethod
    def get_trustee_kind_from_id(id, election_bundle):
        for trustee_kind in election_bundle.trustee_kinds:
            if trustee_kind.trustee_kind_type == "SingleTrusteeModel":
                if trustee_kind.trustee_id == id:
                    return trustee_kind
            else:
                for trustee_id in trustee_kind.trustee_ids:
                    if trustee_id == id:
                        return trustee_kind
        return None

    @staticmethod
    def evaluate_polynomial(a, x, modulus):
        result = 0
        for i, coef in enumerate(a):
            result += coef * pow(x, i)
        return result % modulus



from typing import List

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey, Table
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.ChannelMsgModel import ChannelMsgModel
from belenios.models.belenios.SignedMsgModel import SignedMsgModel

assoc_v_input_secret = Table(
    "assoc_v_input_secret",
    BaseModel.metadata,
    Column("v_input_id", Integer, ForeignKey("v_input_model.id"), primary_key=True),
    Column("secret_id", Integer, ForeignKey("channel_msg_model.id"), primary_key=True)
)
assoc_v_input_coefexps = Table(
    "assoc_v_input_coefexps",
    BaseModel.metadata,
    Column("v_input_id", Integer, ForeignKey("v_input_model.id"), primary_key=True),
    Column("coefexps_id", Integer, ForeignKey("signed_msg_model.id"), primary_key=True)
)
class VInputModel(BaseModel):
    polynomial_id = Column(Integer, ForeignKey('channel_msg_model.id'), nullable=False)
    polynomial: Mapped[ChannelMsgModel] = relationship('ChannelMsgModel', uselist=False, foreign_keys=[polynomial_id])
    secrets: Mapped[List[ChannelMsgModel]] = relationship('ChannelMsgModel', secondary=assoc_v_input_secret)
    coefexps: Mapped[List[SignedMsgModel]] = relationship('SignedMsgModel', secondary=assoc_v_input_coefexps)

    @staticmethod
    def load_from_json(dict):
        t = VInputModel(polynomial=ChannelMsgModel.load_from_json(dict["polynomial"]))
        t.secrets.extend([ChannelMsgModel.load_from_json(s) for s in dict["secrets"]])
        t.coefexps.extend([SignedMsgModel.load_from_json(s) for s in dict["coefexps"]])
        return t



import json

from Crypto.Util.number import getRandomRange

from belenios.dataclass.dataclass import KeysPair
from belenios.models.belenios.CertKeysModel import CertKeysModel
from belenios.models.belenios.PolynomialModel import PolynomialModel
from belenios.models.belenios.ProofModel import ProofModel
from belenios.models.belenios.SignedStringModel import SignedStringModel
from belenios.models.belenios.TrusteePublicKeyModel import TrusteePublicKeyModel
from belenios.models.belenios.VInputModel import VInputModel
from belenios.models.belenios.VOutputModel import VOutputModel
from belenios.utilities.ProofOfKnowledge import ProofOfKnowledge
from belenios.utilities.Utility import Utility
from belenios.utilities.command_line.DbSession import DbSession


class TrusteesKeysProtocol:

    # https://www.belenios.org/specification.pdf#subsection.4.5
    def generate_single_trustee_key(self, election):
        keys_pair = Utility.gen_key_el_gamal(election.group)
        return TrusteesKeysProtocol().generate_trustee_public_key(election, keys_pair)

    def generate_trustee_public_key(self, election, keys_pair):
        # we generate A = g^w using a key pair, where private_key = A and public_key = w
        tmp = Utility.gen_key_el_gamal(election.group)
        w = tmp.private_key
        A = tmp.public_key

        challenge = ProofOfKnowledge.Hpok(election.group, keys_pair.public_key, A) % election.group.q
        response = (w - challenge * keys_pair.private_key)
        trusteePublicKeyModel = TrusteePublicKeyModel()
        trusteePublicKeyModel.public_key = keys_pair.public_key
        trusteePublicKeyModel.pok = ProofModel()
        trusteePublicKeyModel.pok.challenge = challenge
        trusteePublicKeyModel.pok.response = response
        return trusteePublicKeyModel, keys_pair

    # https://www.belenios.org/specification.pdf#subsection.4.5
    def check_proof_trustee_public_key(self, election, trusteePublicKeyModel):
        left = pow(election.group.g, trusteePublicKeyModel.pok.response, election.group.p)
        right = pow(trusteePublicKeyModel.public_key, trusteePublicKeyModel.pok.challenge, election.group.p)
        A = (left * right) % election.group.p
        challenge = ProofOfKnowledge.Hpok(election.group, trusteePublicKeyModel.public_key, A) % election.group.q
        return trusteePublicKeyModel.pok.challenge == challenge

    def derivate_private_seed(self, seed, group):
        sk = int.from_bytes(Utility.hash_sha256_as_bytes(("sk|" + seed).encode()), byteorder='big')
        dk = int.from_bytes(Utility.hash_sha256_as_bytes(("vk|" + seed).encode()), byteorder='big')
        vk = pow(group.g, sk, group.p)
        ek = pow(group.g, dk, group.p)
        return sk, dk, vk, ek

    # https://www.belenios.org/specification.pdf#subsection.4.6
    def generate_perdersen_trustee_keys(self, election):
        seed = Utility.generate_base58_string(22)
        signing_key, decryption_key, verification_key, encryption_key = self.derivate_private_seed(seed, election.group)

        cert_keys = CertKeysModel()
        cert_keys.verification = verification_key
        cert_keys.encryption = encryption_key

        signed_msg = Utility.generate_signed_msg(cert_keys, signing_key, election)
        return signed_msg, seed

    # https://www.belenios.org/specification.pdf#subsubsection.4.6.4
    def generate_perdersen_trustee_polynomial(self, election_bundle, threshold, seed, trustee_id):
        election = election_bundle.election
        signing_key, _, _, encryption_key = self.derivate_private_seed(seed, election.group)
        a_i = [getRandomRange(1, election.group.q - 2) for _ in range(threshold + 1)]
        a_i = [1,2]
        A_i = [pow(election.group.g, a_ij, election.group.p) for a_ij in a_i]
        pedersen_trustee = None
        for pedersen_trustee_iter in election_bundle.trustee_kinds:
            if pedersen_trustee_iter.trustee_kind_type == "PedersenTrusteeModel":
                pedersen_trustee = pedersen_trustee_iter
                break
        if pedersen_trustee is None:
            print("Invalid trustee id")
            return None
        s_ij = [Utility.evaluate_polynomial(a_i, j, election.group.q) for j in range(len(pedersen_trustee.certs) + 1)]
        print("a_i",   a_i)
        print("A_i", A_i)
        print("s_i", s_ij)
        polynomial = PolynomialModel()
        polynomial.trustee_id = trustee_id
        # check A_i or filled with ai0, . . . , ait
        polynomial.polynomial = Utility.generate_channel_msg(json.dumps(a_i), signing_key, encryption_key, election)
        # print("d", Utility.decrypt_channel_msg(polynomial.polynomial, verification_key, decryption_key, election))
        for j, cur_s_ij in enumerate(s_ij[1:]):
            # TODO handle random order adding
            ek_j = pedersen_trustee.certs[j].signed_object.encryption
            polynomial.secrets.append(Utility.generate_channel_msg(str(cur_s_ij), signing_key, ek_j, election))
        signed_string = SignedStringModel()
        signed_string.value = json.dumps(A_i)
        polynomial.coefexps = Utility.generate_signed_msg(signed_string, signing_key, election)

        election_bundle.polynomials.append(polynomial)
        DbSession().session.commit()
        return polynomial

    # https://www.belenios.org/specification.pdf#subsubsection.4.6.5
    def generate_perdersen_trustee_vinput(self, election_bundle, seed, trustee_id):
        election = election_bundle.election
        signing_key, _, _, encryption_key = self.derivate_private_seed(seed, election.group)

        polynomial = None
        for p in election_bundle.polynomials:
            if p.polynomial.recipient == encryption_key and p.trustee_id == trustee_id:
                polynomial = p

        if polynomial is None:
            print("Polynomial not found")
            return None
        v_input = VInputModel()
        v_input.polynomial = polynomial.polynomial
        for p_i in election_bundle.polynomials:
            v_input.secrets.append(p_i.secrets[trustee_id - 1])
            v_input.coefexps.append(p_i.coefexps)

            # for p in p_i.polynomials:
            #     # print(p.trustee_id, p.secrets)
            #     if p.trustee_id == trustee_id:
            #         pass
            #         break

        return v_input

    def check_perdersen_trustee_vinput(self, election_bundle, seed, trustee_id, v_input):
        election = election_bundle.election
        signing_key, decryption_key, verification_key, encryption_key = self.derivate_private_seed(seed, election.group)

        pedersen_trustee = None
        for pedersen_trustee_iter in election_bundle.trustee_kinds:
            if pedersen_trustee_iter.trustee_kind_type == "PedersenTrusteeModel":
                pedersen_trustee = pedersen_trustee_iter
                break
        if pedersen_trustee is None:
            print("Invalid trustee id")
            return False

        for i, coef in enumerate(v_input.coefexps):
            prod = 1
            A_i = json.loads(coef.signed_object.value)
            for k in range(pedersen_trustee.threshold):
                jk = pow(trustee_id, k)
                A_ik = A_i[k]
                # print(jk, A_ik)
                prod *= pow(A_ik, jk, election_bundle.election.group.p)
                prod = prod % election_bundle.election.group.p
            vk_j = pedersen_trustee.certs[i].signed_object.verification
            s_ij = int(Utility.decrypt_channel_msg(v_input.secrets[i], vk_j, decryption_key,
                                               election).signed_object.value)

            g_s_ij = pow(election_bundle.election.group.g, s_ij, election_bundle.election.group.p)
            # print(g_s_ij, prod)
            if g_s_ij != prod:
                print("Incorrect product", i, k, g_s_ij, prod)
                return False
        return True

    # https://www.belenios.org/specification.pdf#subsubsection.4.6.6
    def generate_perdersen_trustee_voutput(self, election_bundle, seed, trustee_id, v_input):
        election = election_bundle.election
        signing_key, decryption_key, verification_key, encryption_key = self.derivate_private_seed(seed, election.group)

        pedersen_trustee = None
        for pedersen_trustee_iter in election_bundle.trustee_kinds:
            if pedersen_trustee_iter.trustee_kind_type == "PedersenTrusteeModel":
                pedersen_trustee = pedersen_trustee_iter
                break
        if pedersen_trustee is None:
            print("Invalid trustee id")
            return None

        private_key = 0
        for i, s_i in enumerate(v_input.secrets):
            # check sign with verification key
            vk_i = pedersen_trustee.certs[i].signed_object.verification
            decrypted = Utility.decrypt_channel_msg(s_i, vk_i, decryption_key, election)
            # print("d", decrypted)
            # print("d", decrypted.signed_object.value)

            private_key += int(decrypted.signed_object.value)
            # private_key = private_key % election.group.q
        keys_pair = KeysPair(pow(election.group.g, private_key, election.group.p), private_key)
        v_output = VOutputModel()
        v_output.private_key = Utility.generate_channel_msg(str(private_key), signing_key, encryption_key,
                                                            election)
        v_output.trustee_public_key, _ = TrusteesKeysProtocol().generate_trustee_public_key(election, keys_pair)
        return v_output, private_key
    def compute_election_public_key(self,election_bundle):
        result = 1
        for trustee_kind in election_bundle.trustee_kinds:
            if trustee_kind.trustee_kind_type == "SingleTrusteeModel":
                result = (result * trustee_kind.trustee_public_key.public_key) % election_bundle.election.group.p
            else:
                for coefexps in trustee_kind.coefexps:
                    A_i = json.loads(coefexps.signed_object.value)
                    result = (result * A_i[0]) % election_bundle.election.group.p
        #           TODO: Check modulo
        return result



from sqlalchemy import Column

from belenios.custom_types.belenios.HType import HType
from belenios.custom_types.belenios.IType import IType
from belenios.models.BaseModel import BaseModel


class SizedEncryptedTallyModel(BaseModel):
    num_tallied = Column(IType, nullable=True)
    total_weight = Column(IType, nullable=True)
    encrypted_tally = Column(HType, nullable=True)

    def __str__(self):
        return (
            f"SizedEncryptedTallyModel(\n"
            f"    num_tallied={self.num_tallied},\n"
            f"    total_weight={self.total_weight},\n"
            f"    encrypted_tally={self.encrypted_tally}\n"
            f")"
        )


from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship

from belenios.custom_types.belenios.PublicKeyType import PublicKeyType
from belenios.models.BaseModel import BaseModel


class TrusteePublicKeyModel(BaseModel):
    proof_id = Column(Integer, ForeignKey('proof_model.id'), nullable=False)
    pok = relationship('ProofModel')
    public_key = Column(PublicKeyType, nullable=False)

    def to_json(self):
        return {
            "proof_id": self.proof_id,
            "public_key": self.public_key,
            "pok": self.pok.to_json() if self.pok else None
        }


from belenios.models.belenios.DecryptionFactorRowModel import DecryptionFactorRowModel
from belenios.models.belenios.OwnedPartialDecryptionModel import OwnedPartialDecryptionModel
from belenios.models.belenios.PartialDecryptionModel import PartialDecryptionModel
from belenios.utilities.TrusteesKeysProtocol import TrusteesKeysProtocol
from belenios.utilities.Utility import Utility
from belenios.utilities.command_line.DbSession import DbSession


class TrusteeCore:

    @staticmethod
    def generates_single_trustee_public_key(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        return TrusteesKeysProtocol().generate_single_trustee_key(election_bundle.election)

    @staticmethod
    def generates_perdersen_trustee_keys(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        return TrusteesKeysProtocol().generate_perdersen_trustee_keys(election_bundle.election)

    @staticmethod
    def generates_perdersen_trustee_polynomial(election_uuid, threshold, seed, trustee_id):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        return TrusteesKeysProtocol().generate_perdersen_trustee_polynomial(election_bundle, threshold, seed,
                                                                            trustee_id)
    @staticmethod
    def generates_perdersen_trustee_vinput(election_uuid, seed, trustee_id):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        return TrusteesKeysProtocol().generate_perdersen_trustee_vinput(election_bundle, seed,trustee_id)

    @staticmethod
    def check_perdersen_trustee_vinput(election_uuid, seed, trustee_id, vinput):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        return TrusteesKeysProtocol().check_perdersen_trustee_vinput(election_bundle, seed, trustee_id, vinput)

    @staticmethod
    def generates_perdersen_trustee_voutput(election_uuid, seed, trustee_id, v_input):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        return TrusteesKeysProtocol().generate_perdersen_trustee_voutput(election_bundle, seed, trustee_id,
                                                                         v_input)

    @staticmethod
    def partial_decrypt_encrypted_ballot(election_uuid, private_key, trustee_id):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        owned_partial_decryption = OwnedPartialDecryptionModel()
        owned_partial_decryption.trustee_id = trustee_id
        owned_partial_decryption.partial_decryption = PartialDecryptionModel()

        p = election_bundle.election.group.p
        for encrypted_tally_row in election_bundle.tally.encrypted_tally_rows:
            # print("dd")
            decryption_factor_row = DecryptionFactorRowModel()
            # print("dd", encrypted_tally_row)
            # print("dd", encrypted_tally_row.ciphertexts)

            for ciphertext in encrypted_tally_row.ciphertexts:
                # TODO: check mod q or g  and -1
                decryption_factor = pow(ciphertext.alpha, -1 * private_key, p)
                # a = EncryptedText([(ciphertext.alpha, ciphertext.beta)])
                # print("test", ElGamal().decrypt_from_int(a, private_key ,election_bundle.election.group, True, decryption_factor))
                decryption_factor_row.decryption_factors.append(decryption_factor)
            #     TODO: make decryption_proof_row
            owned_partial_decryption.partial_decryption.decryption_factor_rows.append(decryption_factor_row)
        election_bundle.owned_partial_decryptions.append(owned_partial_decryption)
        DbSession().session.commit()
        return owned_partial_decryption



from sqlalchemy.types import TypeDecorator, VARCHAR


class HType(TypeDecorator):
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            # Ensure value is a string and convert it to lowercase
            value = str(value).lower()
            # Validate hash length (SHA256 produces 64-character hashes)
            if len(value) != 64:
                raise ValueError("Hash value must be 64 characters long")
        return value

    def process_result_value(self, value, dialect):
        return value

    def process_literal_param(self, value, dialect):
        return value



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.GType import GType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.EncryptedDataModel import EncryptedDataModel


class EncryptedMsgModel(BaseModel):
    alpha = Column(GType, nullable=False)
    beta = Column(GType, nullable=False)
    data_id = Column(Integer, ForeignKey('encrypted_data_model.id'), nullable=False)
    data: Mapped[EncryptedDataModel] = relationship('EncryptedDataModel', uselist=False)

    @staticmethod
    def load_from_json(dict):
        return EncryptedMsgModel(alpha=int(dict["alpha"]), beta=int(dict["beta"]), data=EncryptedDataModel.load_from_json(dict["data"]))



from sqlalchemy import Column

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class EncryptedDataModel(BaseModel):
    ciphertext = Column(StringType, nullable=False)
    tag = Column(StringType, nullable=False)
    iv = Column(StringType, nullable=False)

    @staticmethod
    def load_from_json(dict):
        return EncryptedDataModel(ciphertext=dict["ciphertext"], tag=dict["tag"], iv=dict["iv"])



from typing import List

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.ResultRowModel import ResultRowModel


class ResultModel(BaseModel):
    result_rows: Mapped[List[ResultRowModel]] = relationship("ResultRowModel", uselist=True,
                                                             cascade="all, delete-orphan",
                                                             single_parent=True)

    def __str__(self):
        result_rows_info = ', '.join(str(result_row) for result_row in self.result_rows)
        return (
            f"ResultModel(\n"
            f"    result_rows=[{result_rows_info}]\n"
            f")"
        )


from sqlalchemy import Column

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class SetupDataModel(BaseModel):
    election = Column(StringType, nullable=True)
    trustees = Column(StringType, nullable=True)
    credentials = Column(StringType, nullable=True)

    def __str__(self):
        return (
            f"SetupDataModel(\n"
            f"    election={self.election},\n"
            f"    trustees={self.trustees},\n"
            f"    credentials={self.credentials}\n"
            f")"
        )


from typing import List

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.GType import GType
from belenios.custom_types.belenios.StringType import StringType
from belenios.custom_types.belenios.UuidType import UuidType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.AnswerModel import AnswerModel
from belenios.models.belenios.SignatureModel import SignatureModel
from belenios.utilities.Utility import Utility


class BallotModel(BaseModel):
    tally_id = Column(Integer, ForeignKey('tally_model.id'), nullable=True)
    election_uuid = Column(UuidType, nullable=True)
    election_hash = Column(StringType, nullable=False)
    credential = Column(GType, nullable=False)  # A ballot references in its credential member the public \n
    # credential S = gsecret(c,s) (c being the secret credential, s being the associated salt) of the voter
    answers: Mapped[List[AnswerModel]] = relationship('AnswerModel', uselist=True, cascade="all, delete-orphan",
                                                         single_parent=True)
    # TODO: add credential_id
    signature: Mapped[SignatureModel] = relationship('SignatureModel', uselist=False, single_parent=True)
    # TODO: change to_hash_sha256 to respect To compute the hash used in signatures, the ballot without the signature
    #  field is first serialized as a JSON compact string, where object fields are ordered as specified in this document
    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"election_uuid={self.election_uuid}" + delimiter +
                f"election_hash={self.election_hash}" + delimiter +
                f"credential={self.credential}" + delimiter +
                f"answers={[answer.to_hash_sha256() for answer in self.answers]}" + delimiter +
                f"signature=EXCLUDED"  # The signature needs a credential c and uses the hash of the surrounding
            # ballot (without the signature field)
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash
    def __str__(self):
        answers_info = ', '.join(str(answer) for answer in self.answers)
        return (
            f"BallotModel(\n"
            f"    election_uuid={self.election_uuid},\n"
            f"    election_hash={self.election_hash},\n"
            f"    credential={self.credential},\n"
            f"    answers=[{answers_info}],\n"
            f"    signature={self.signature}\n"
            f")"
        )


from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey, Table
from sqlalchemy.orm import relationship

from belenios.models.BaseModel import BaseModel
from belenios.utilities.Utility import Utility

assoc_i_proof_and_proofs = Table(
    "assoc_i_proof_and_proofs",
    BaseModel.metadata,
    Column("i_proof_model_id", ForeignKey("i_proof_model.id"), primary_key=True),
    Column("proof_model_id", ForeignKey("proof_model.id"), primary_key=True),
)


class IProofModel(BaseModel):
    i_proof_model_id = Column(Integer, ForeignKey('i_proof_model.id'))
    proofs = relationship("ProofModel", secondary=assoc_i_proof_and_proofs)

    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"id={self.id}" + delimiter +
                f"proofs={[proof.to_hash_sha256() for proof in self.proofs]}" + delimiter
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash



from Crypto.Util.number import inverse

from belenios.models.belenios.AnswerHModel import AnswerHModel
from belenios.models.belenios.AnswerNhModel import AnswerNhModel
from belenios.models.belenios.CiphertextModel import CiphertextModel
from belenios.models.belenios.IProofModel import IProofModel
from belenios.models.belenios.ProofModel import ProofModel
from belenios.utilities.ElGamal import ElGamal
from belenios.utilities.ProofOfKnowledge import ProofOfKnowledge
from belenios.utilities.Utility import Utility


class AnswerUtility:
    @staticmethod
    def gen_S0(election_bundle, secret_c):
        return election_bundle.setup_data.election + "|" + str(
            pow(election_bundle.election.group.g, secret_c, election_bundle.election.group.q))

    @staticmethod
    def gen_S(election_bundle, secret_c, ciphertext_model_arr, ):
        a = [str(c.alpha) + "," + str(c.beta) for c in ciphertext_model_arr]
        b = ','.join(a)
        return AnswerUtility.gen_S0(election_bundle, secret_c) + '|' + b

    @staticmethod
    def compute_iprove(election, S, choices_for_a_question, encrypted_text_and_k, id_answer, answer):
        encryption = answer.choices[id_answer]
        alpha = encrypted_text_and_k.chunk[0][0]
        beta = encrypted_text_and_k.chunk[0][1]
        k = encrypted_text_and_k.chunk[0][2]
        proofs = []
        Ajs = []
        Bjs = []
        sum_challenge = 0
        # 1. for j ̸= i:
        for j, choice in enumerate(choices_for_a_question):
            if j == id_answer:
                continue
            pi_proof = ProofModel()
            pi_proof.challenge = Utility.generate_number_in_q(election.group)
            pi_proof.response = Utility.generate_number_in_q(election.group)
            proofs.append(pi_proof)
            sum_challenge += pi_proof.challenge
            Aj = (pow(election.group.g, pi_proof.response, election.group.p) * pow(alpha, pi_proof.challenge,
                                                                                   election.group.p)) % election.group.p
            Ajs.append(Aj)
            inv_g_Mi = inverse(pow(election.group.g, choice, election.group.p), election.group.p)
            beta_factor = (beta * inv_g_Mi) % election.group.p
            Bj = (pow(election.public_key, pi_proof.response, election.group.p) * pow(beta_factor, pi_proof.challenge,
                                                                                      election.group.p)) % election.group.p
            Bjs.append(Bj)
        # 2. πi is created as follows:
        # we generate A = g^w using a key pair, where private_key = A and public_key = w
        tmp = Utility.gen_key_el_gamal(election.group)
        w = tmp.private_key
        Ai = tmp.public_key

        Ajs.insert(id_answer, Ai)
        Bj = pow(election.public_key, w, election.group.p)
        Bjs.insert(id_answer, Bj)

        hash = ProofOfKnowledge.Hiprove(S, encryption, Ajs, Bjs)
        pi_proof = ProofModel()
        pi_proof.challenge = hash - (sum_challenge % election.group.q)
        pi_proof.response = (w - k * pi_proof.challenge) % election.group.q
        proofs.insert(id_answer, pi_proof)
        iproof = IProofModel()
        iproof.proofs = proofs
        return iproof

    @staticmethod
    def check_iprove(election, S, choices_for_a_question, id_answer, id_choice, answer):
        pi_proofs = answer.individual_proofs[id_choice].proofs
        encryption = answer.choices[id_choice]
        alpha = encryption.alpha
        beta = encryption.beta
        Ajs = []
        Bjs = []
        sum_challenge = 0
        # 1. for j ̸= i:
        for j, (choice, pi_proof) in enumerate(zip(choices_for_a_question, pi_proofs)):
            sum_challenge += pi_proof.challenge
            Aj = (pow(election.group.g, pi_proof.response, election.group.p) * pow(alpha, pi_proof.challenge,
                                                                                   election.group.p)) % election.group.p
            Ajs.append(Aj)
            inv_g_Mi = inverse(pow(election.group.g, choice, election.group.p), election.group.p)
            beta_factor = (beta * inv_g_Mi) % election.group.p
            Bj = (pow(election.public_key, pi_proof.response, election.group.p) * pow(beta_factor, pi_proof.challenge,
                                                                                      election.group.p)) % election.group.p
            Bjs.append(Bj)

        hash = ProofOfKnowledge.Hiprove(S, encryption, Ajs, Bjs)
        sum_challenge -= pi_proofs[id_choice].challenge
        sum_challenge %= election.group.q
        return hash == sum_challenge

    @staticmethod
    def compute_individuals_proofs(election_bundle, secret, answer, choices_for_a_question, encrypted_text_and_k_arr):
        S = AnswerUtility.gen_S(election_bundle, secret, answer.choices)
        for id_answer, encrypted_text_and_k in enumerate(encrypted_text_and_k_arr):
            individual_proof = AnswerUtility.compute_iprove(election_bundle.election, S, choices_for_a_question,
                                                            encrypted_text_and_k, id_answer, answer)
            answer.individual_proofs.append(individual_proof)

    # https://www.belenios.org/specification.pdf#subsubsection.4.11.1
    @staticmethod
    def generate_answer_for_h_question(election_uuid, blank_allowed, choices_for_a_question, secret):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        election = election_bundle.election
        if election is None:
            print("Invalid election")
            return None
        answer = AnswerHModel()
        encrypted_text_and_k_arr = []
        for id_answer, choice in enumerate(choices_for_a_question):
            m = choice
            encrypted_text_and_k = ElGamal().encrypt_from_int(m, election.public_key, election.group, save_k=True,
                                                              exponential=True)
            if len(encrypted_text_and_k.chunk) != 1:
                print("Error during encryption of the answer")
                return None

            encrypted_text_and_k_arr.append(encrypted_text_and_k)
            ciphertext_model = CiphertextModel()
            ciphertext_model.alpha = encrypted_text_and_k.chunk[0][0]
            ciphertext_model.beta = encrypted_text_and_k.chunk[0][1]
            answer.choices.append(ciphertext_model)
        AnswerUtility.compute_individuals_proofs(election_bundle, secret, answer, choices_for_a_question,
                                                 encrypted_text_and_k_arr)
        # TODO: implement iproov and ballot proof
        # if blank_allowed:
        #     print("blank not implemented yet")
        #     return None
        # else:
        #     # When a blank vote is not allowed, overall_proof proves that M ∈ [min . . . max] and is computed
        #     # by running iprove(S, R, M − min, min, . . . , max) where R is the sum of the r used in ciphertexts,
        #     # and M the sum of the m. There is no blank_proof.
        #     pass
        answer.overall_proof = ProofModel()
        answer.overall_proof.challenge = -1
        answer.overall_proof.response = -1

        return answer

    # https://www.belenios.org/specification.pdf#subsubsection.4.11.2
    @staticmethod
    def generate_answer_for_nh_question(election_uuid, choices, private_credential, public_credential, salt):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        election = election_bundle.election
        if election is None:
            print("Invalid election")
            return None

        # TODO: implement AnswerNh
        print("not implemented yet")
        return None

        answer = AnswerNhModel()
        answer.choices = CiphertextModel()
        xi = str([1 if answerStr == answerString else 0 for answerStr in question.answers]).encode()

        encrypted_text_and_k = ElGamal().encrypt(xi, election.public_key, election.group, save_k=True)
        if len(encrypted_text_and_k.chunk) != 1:
            print("Error during encryption of the answer")
            return None
        r = encrypted_text_and_k.chunk[0][0]
        answer.choices.alpha = encrypted_text_and_k.chunk[0][1]
        answer.choices.beta = encrypted_text_and_k.chunk[0][2]

        # IF multiple chunk is possible
        # if len(encrypted_text_and_k.chunk) > 0:
        #     print("Error during encryption of the answer")
        #     return None
        # r = [chunk[0] for chunk in encrypted_text_and_k.chunk]
        # answer.choices.alpha = [chunk[1] for chunk in encrypted_text_and_k.chunk]
        # answer.choices.beta = [chunk[2] for chunk in encrypted_text_and_k.chunk]
        # TODO: Check here
        # secret_c = Utility.derive_secret_exponent(credential, salt, election)
        # S0 = AnswerUtility.gen_S0(election_bundle, secret_c)

        S = ballot.credential
        # we generate A = g^w using a key pair, where private_key = A and public_key = w
        tmp = Utility.gen_key_el_gamal(election.group)
        w = tmp.private_key
        A = tmp.public_key

        challenge = ProofOfKnowledge.Hraweg(S, election, answer, A) % election.group.q
        response = (w - r * challenge)
        answer.proof = ProofModel()
        answer.proof.challenge = challenge
        answer.proof.response = response
        ballot.answers.append(answer)
        return ballot

    @staticmethod
    def check_single_nh_answer_proof(election_bundle, answer, credential):
        # TODO: Check here
        # secret_c = Utility.derive_secret_exponent(credential, salt, election_bundle.election)
        # S0 = AnswerUtility.gen_S0(election_bundle, secret_c)
        S = credential
        left = pow(election_bundle.election.group.g, answer.proof.response, election_bundle.election.group.p)
        right = pow(answer.choices.alpha, answer.proof.challenge, election_bundle.election.group.p)
        A = (left * right) % election_bundle.election.group.q
        challenge = ProofOfKnowledge.Hraweg(S, election_bundle.election, answer, A) % election_bundle.election.group.q
        return answer.proof.challenge == challenge

    @staticmethod
    def check_sign_ballot(election_uuid, ballot):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        left = pow(election_bundle.election.group.g, ballot.signature.proof.response, election_bundle.election.group.p)
        right = pow(ballot.credential, ballot.signature.proof.challenge, election_bundle.election.group.p)
        A = (left * right) % election_bundle.election.group.p
        H = ballot.signature.hash
        challenge = ProofOfKnowledge.Hsignature(H, A) % election_bundle.election.group.q
        return ballot.signature.proof.challenge == challenge

    @staticmethod
    def check_ballot(election_uuid, ballot, secret, choices_for_all_questions):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        # check the signature
        if not AnswerUtility.check_sign_ballot(election_uuid, ballot):
            return False

        # check iproof
        for id_answer, answer in enumerate(ballot.answers):
            S = AnswerUtility.gen_S(election_bundle, secret, answer.choices)
            for id_choice, choice in enumerate(answer.choices):
                if AnswerUtility.check_iprove(election_bundle.election, S, choices_for_all_questions[id_answer],
                                              id_answer, id_choice, answer):
                    print("invalid proof")
                    return False

        # TODO: check iproof
        # check blank_proof
        # check overall_proof
        # for answer in ballot.answers:
        #     if not AnswerUtility.check_single_nh_answer_proof(election_bundle, answer, ballot.credential):
        #         return False
        return True



# from typing import List
#
# from sqlalchemy import Column
# from sqlalchemy import Integer, ForeignKey
# from sqlalchemy.orm import relationship, Mapped
#
# from belenios.models.BaseModel import BaseModel
# from belenios.models.belenios.TrusteeKindModel import TrusteeKindModel
#
#
# class TrusteeKindRowModel(BaseModel):
#     election_bundle_id = Column(Integer, ForeignKey('election_bundle_model.id'), nullable=False)
#     trustee_kinds: Mapped[List[TrusteeKindModel]] = relationship('TrusteeKindModel', uselist=True,
#                                                                  cascade="all, delete-orphan",
#                                                                  single_parent=True)
#
#     def __init__(self, *args, **kwargs):
#         if 'trustee_kinds' not in kwargs:
#             kwargs['trustee_kinds'] = []
#         super().__init__(*args, **kwargs)



from cement import Controller
from cement.utils.version import get_version_banner

from belenios.core.command_line.version import get_version

VERSION_BANNER = """
This is a proof of concept of the Belenios election protocol. %s
%s
""" % (get_version(), get_version_banner())


class DefaultCmd(Controller):
    class Meta:
        label = 'base'

        # base_controller = True

        # text displayed at the top of --help output
        description = 'This is a proof of concept of the Belenios election protocol.'

        # text displayed at the bottom of --help output
        epilog = 'Usage: belenios'

        # controller level arguments. ex: 'belenios --version'
        arguments = [
            ### add a version banner
            (['-v', '--version'],
             {'action': 'version',
              'version': VERSION_BANNER}),
        ]

    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()



import json

from sqlalchemy.types import TypeDecorator, VARCHAR


class ArrayOfGsType(TypeDecorator):
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            # Serialize the array of ints to strings JSON
            if not isinstance(value, list):
                value = [value]
            for i, v in enumerate(value):
                value[i] = str(v)
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            # Deserialize the JSON to an array of ints
            value = [int(v) for v in json.loads(value)]
        return value

    def process_literal_param(self, value, dialect):
        return value



from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.IType import IType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.PartialDecryptionModel import PartialDecryptionModel
from belenios.models.belenios.TrusteeKindModel import TrusteeKindModel


class OwnedPartialDecryptionModel(BaseModel):
    election_bundle_id: int = Column(Integer, ForeignKey('election_bundle_model.id'), nullable=False)
    trustee_id = Column(IType, nullable=False)
    partial_decryption: Mapped[PartialDecryptionModel] = relationship('PartialDecryptionModel',
                                                                      uselist=False)  # payload : H
    partial_decryption_id = Column(Integer, ForeignKey('partial_decryption_model.id'))

    def __str__(self):
        return (
            f"OwnedPartialDecryptionModel(\n"
            # f"    election_bundle_id={self.election_bundle_id},\n"
            f"    trustee_id={self.trustee_id},\n"
            f"    trustee_kind={self.trustee_kind},\n"
            f"    partial_decryption={self.partial_decryption},\n"
            # f"    partial_decryption_id={self.partial_decryption_id}\n"
            f")"
        )



from typing import List

from sqlalchemy import Column, Table, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.IType import IType
from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel

# note for a Core table, we use the sqlalchemy.Column construct,
# not sqlalchemy.orm.mapped_column
assoc_elections_bundle_and_voters = Table(
    "assoc_elections_bundle_and_voters",
    BaseModel.metadata,
    Column("election_bundle_model_id", ForeignKey("election_bundle_model.id"), primary_key=True),
    Column("voter_model_id", ForeignKey("voter_model.id"), primary_key=True),
)


class VoterModel(BaseModel):
    email = Column(StringType, nullable=False, unique=True)
    weight = Column(IType, nullable=False)
    election_bundles: Mapped[List['ElectionBundleModel']] = relationship(secondary=assoc_elections_bundle_and_voters,
                                                                         back_populates="voters")
    def __str__(self):
        return (
            f"VoterModel(\n"
            f"    email={self.email},\n"
            f"    weight={self.weight},\n"
            # f"    election_bundles={self.election_bundles}\n"
            f")"
        )


from abc import abstractmethod

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class SignedObjectModel(BaseModel):
    # __abstract__ = True
    type = Column(StringType, nullable=False)
    __mapper_args__ = {
        "polymorphic_identity": "SignedObject",
        "polymorphic_on": "type",
    }

    @abstractmethod
    def to_hash_sha256(self):
        pass

    @staticmethod
    def load_from_json(dict):
        if dict['type'] == "RawCoefexpsModel":
            from belenios.models.belenios.RawCoefexpsModel import RawCoefexpsModel
            return RawCoefexpsModel.load_from_json(dict)
        elif dict['type'] == "SignedStringModel":
            from belenios.models.belenios.SignedStringModel import SignedStringModel
            return SignedStringModel.load_from_json(dict)
        else:
            print("Type not implemented", dict['type'])
            return None



class UserSession:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._initialize()
        return cls._instance

    def _initialize(self):
        self.user = None



from sqlalchemy import Column, Integer, ForeignKey

from belenios.custom_types.belenios.GType import GType
from belenios.models.belenios.SignedObjectModel import SignedObjectModel


class CertKeysModel(SignedObjectModel):
    id = Column(Integer, ForeignKey('signed_object_model.id'), primary_key=True)
    verification = Column(GType)
    encryption = Column(GType)
    __mapper_args__ = {
        "polymorphic_identity": "CertKeysModel",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'CertKeysModel'

    def __str__(self):
        return (
            f"CertKeysModel(\n"
            f"    verification={self.verification},\n"
            f"    encryption={self.encryption}\n"
            f")"
        )
    def to_json(self):
        return {
            "verification": self.verification,
            "encryption": self.encryption,
        }


from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.IType import IType
from belenios.custom_types.belenios.StringType import StringType
from belenios.models.BaseModel import BaseModel


class CredentialModel(BaseModel):
    index = Column(IType, nullable=False)
    public_credential = Column(StringType, nullable=False)
    weight = Column(IType, nullable=True)
    email = Column(StringType, nullable=False)
    election_bundle_id = Column(Integer, ForeignKey('election_bundle_model.id'), nullable=False)
    election_bundle: Mapped['ElectionBundleModel'] = relationship("ElectionBundleModel",
                                                                  back_populates="public_credentials")

    def to_json(self):
        return {
            'index': self.index,
            'public_credential': self.public_credential,
            'weight': self.weight,
            'email': self.email
        }
    def __str__(self):
        return (
            f"CredentialModel(\n"
            f"    index={self.index},\n"
            f"    public_credential={self.public_credential},\n"
            f"    weight={self.weight},\n"
            f"    email={self.email},\n"
            # f"    election_bundle_id={self.election_bundle_id},\n"
            # f"    election_bundle={self.election_bundle}\n"
            f")"
        )


from sqlalchemy import Column, Integer, Enum, ForeignKey

from belenios.custom_types.belenios.HType import HType
from belenios.custom_types.belenios.IType import IType
from belenios.models.BaseModel import BaseModel
from belenios.custom_types.belenios.EventType import EventType


class EventModel(BaseModel):
    election_bundle_id: int = Column(Integer, ForeignKey('election_bundle_model.id'), nullable=False)
    parent = Column(HType, nullable=True)
    height = Column(IType, nullable=False)
    type = Column(Enum(EventType), nullable=False)
    payload = Column(HType, nullable=True)

    def __str__(self):
        return (
            f"EventModel(\n"
            # f"    election_bundle_id={self.election_bundle_id},\n"
            f"    parent={self.parent},\n"
            f"    height={self.height},\n"
            f"    type={self.type},\n"
            f"    payload={self.payload}\n"
            f")"
        )


from typing import List

from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.BallotModel import BallotModel
from belenios.models.belenios.EncryptedTallyRowModel import EncryptedTallyRowModel
from belenios.models.belenios.SizedEncryptedTallyModel import SizedEncryptedTallyModel


class TallyModel(BaseModel):
    ballots: Mapped[List[BallotModel]] = relationship('BallotModel', uselist=True, cascade="all, delete-orphan",
                                                      single_parent=True)
    election_bundle_id = Column(Integer, ForeignKey('election_bundle_model.id'), nullable=False, unique=True)
    election_bundle = relationship('ElectionBundleModel', back_populates='tally')
    encrypted_tally_rows: Mapped[List[EncryptedTallyRowModel]] = relationship("EncryptedTallyRowModel", uselist=True,
                                                                         cascade="all, delete-orphan",
                                                                         single_parent=True, back_populates="tally")
    sized_encrypted_tally_id = Column(Integer, ForeignKey('sized_encrypted_tally_model.id'), nullable=False)
    sized_encrypted_tally: Mapped[SizedEncryptedTallyModel] = relationship('SizedEncryptedTallyModel')

    def __str__(self):
        encrypted_tally_rows_info = ', '.join(
            str(encrypted_tally_row) for encrypted_tally_row in self.encrypted_tally_rows)

        return (
            f"TallyModel(\n"
            f"    ballots={self.ballots},\n"
            # f"    election_bundle_id={self.election_bundle_id},\n"
            # f"    election_bundle={self.election_bundle},\n"
            f"    encrypted_tally_rows=[{encrypted_tally_rows_info}],\n"
            f"    sized_encrypted_tally={self.sized_encrypted_tally}\n"
            f")"
        )



from enum import Enum


class EventType(Enum):
    Setup = 'Setup'
    Ballot = 'Ballot'
    EndBallots = 'EndBallots'
    EncryptedTally = 'EncryptedTally'
    Shuffle = 'Shuffle'
    EndShuffles = 'EndShuffles'
    PartialDecryption = 'PartialDecryption'
    Result = 'Result'



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.models.belenios.AnswerModel import AnswerModel
from belenios.utilities.Utility import Utility


class AnswerNhModel(AnswerModel):
    id = Column(Integer, ForeignKey('answer_model.id'), primary_key=True)
    choices: Mapped["CiphertextModel"] = relationship('CiphertextModel', uselist=False)
    choices_id: int = Column(Integer, ForeignKey('ciphertext_model.id'), nullable=False)
    proof_id: int = Column(Integer, ForeignKey('proof_model.id'), nullable=False)
    proof: Mapped['ProofModel'] = relationship('ProofModel')
    __mapper_args__ = {
        "polymorphic_identity": "AnswerNhModel",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'AnswerNhModel'

    def __str__(self):
        return (
            f"AnswerNhModel(id={self.id}, "
            f"choices_id={self.choices_id}, "
            f"choices={self.choices}, "
            f"proof_id={self.proof_id},"
            f"proof={self.proof})"
        )

    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"id={self.id}" + delimiter +
                f"choices={self.choice.to_hash_sha256() }" + delimiter +
                f"individual_proofs={[proof.to_hash_sha256() for proof in self.individual_proofs]}" + delimiter +
                f"proof={self.proof.to_hash_sha256()}"
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash



from sqlalchemy import Column, Integer, ForeignKey

from belenios.models.belenios.QuestionModel import QuestionModel


class QuestionNhModel(QuestionModel):
    id = Column(Integer, ForeignKey('question_model.id'), primary_key=True)
    __mapper_args__ = {
        "polymorphic_identity": "QuestionNhModel",
    }

    # Override constructor to set 'type' to 'Homomorphic'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'QuestionNhModel'

    def __str__(self):
        return (
            f"QuestionNhModel(\n"
            f"    id={self.id},\n"
            f"    type={self.type},\n"
            f"    question={self.question},\n"
            f"    answers={self.answers},\n"
            f"    extra={self.extra}\n"
            f")"
        )


from typing import List

from sqlalchemy import Column
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.IType import IType
from belenios.custom_types.belenios.PublicKeyType import PublicKeyType
from belenios.custom_types.belenios.StringType import StringType
from belenios.custom_types.belenios.UuidType import UuidType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.GroupModel import GroupModel
from belenios.models.belenios.QuestionModel import QuestionModel
from belenios.utilities.Utility import Utility


class ElectionModel(BaseModel):
    version = Column(IType, nullable=False)
    description = Column(StringType, nullable=False)
    name = Column(StringType, nullable=False)
    group: Mapped[GroupModel] = relationship('GroupModel', back_populates='election',
                                                            uselist=False, cascade="all, delete-orphan",
                                                            single_parent=True)
    public_key = Column(PublicKeyType, nullable=True)
    questions: Mapped[List["QuestionModel"]] = relationship('QuestionModel', back_populates='election',
                                                            uselist=True, cascade="all, delete-orphan",
                                                            single_parent=True)
    uuid = Column(UuidType, nullable=True)
    administrator = Column(StringType, nullable=True)
    credential_authority = Column(StringType, nullable=True)
    election_bundle: Mapped["ElectionBundleModel"] = relationship('ElectionBundleModel')
    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"version={self.version}" + delimiter +
                f"description={self.description}" + delimiter +
                f"name={self.name}" + delimiter +
                f"group={self.group.to_hash_sha256()}" + delimiter +
                f"public_key={self.public_key}" + delimiter +
                f"questions={[question.to_hash_sha256() for question in self.questions]}" + delimiter +
                f"uuid={self.uuid}" + delimiter +
                f"administrator={self.administrator}" + delimiter +
                f"credential_authority={self.credential_authority}"
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash

    def __str__(self):
        return (
            f"ElectionModel(\n"
            f"    version={self.version},\n"
            f"    description={self.description},\n"
            f"    name={self.name},\n"
            f"    group={self.group},\n"
            f"    public_key={self.public_key},\n"
            f"    questions={[str(question) for question in self.questions]},\n"
            f"    uuid={self.uuid},\n"
            f"    administrator={self.administrator},\n"
            f"    credential_authority={self.credential_authority},\n"
            # f"    election_bundle={self.election_bundle}\n"
            f")"
        )


from sqlalchemy import Column, Integer

from belenios.custom_types.belenios.IType import IType
from belenios.models.BaseModel import BaseModel


class HeaderModel(BaseModel):
    version = Column(IType, nullable=False)
    timestamp = Column(IType, nullable=False)



from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey, Table
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel

assoc_decryption_proof_row_and_ciphertext = Table(
    "assoc_decryption_proof_row_and_ciphertext",
    BaseModel.metadata,
    Column("decryption_proof_row_model_id", ForeignKey("decryption_proof_row_model.id"), primary_key=True),
    Column("proof_model_id", ForeignKey("proof_model.id"), primary_key=True),
)


class DecryptionProofRowModel(BaseModel):
    partial_decryption_model_id = Column(Integer, ForeignKey('partial_decryption_model.id'))
    proofs = relationship("ProofModel", secondary=assoc_decryption_proof_row_and_ciphertext)



from typing import List

from belenios.models.belenios.DecryptionProofRowModel import DecryptionProofRowModel
from sqlalchemy.orm import relationship, Mapped

from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.DecryptionFactorRowModel import DecryptionFactorRowModel


class PartialDecryptionModel(BaseModel):
    decryption_factor_rows: Mapped[List[DecryptionFactorRowModel]] = relationship("DecryptionFactorRowModel",
                                                                                  uselist=True,
                                                                                  cascade="all, delete-orphan",
                                                                                  single_parent=True)
    decryption_proofs_rows: Mapped[List[DecryptionProofRowModel]] = relationship("DecryptionProofRowModel",
                                                                                 uselist=True,
                                                                                 cascade="all, delete-orphan",
                                                                                 single_parent=True)



from dataclasses import asdict

from Crypto.Random import random

from belenios.dataclass.dataclass import PrivateCredential
from belenios.models.belenios.CredentialModel import CredentialModel
from belenios.models.belenios.SaltModel import SaltModel
from belenios.models.belenios.VoterModel import VoterModel
from belenios.utilities.Utility import Utility
from belenios.utilities.command_line.DbSession import DbSession


class CredentialAuthorityCore:

    @staticmethod
    def generate_secret_credential(index_already_used):
        # Generate random prefix of 14 characters from Base58 alphabet
        prefix = Utility.generate_base58_string(14)

        # Generate random index NNNN
        index = random.randint(0, 9999)
        for i in range(10000):
            if index in index_already_used:
                index = random.randint(0, 9999)
            else:
                break

        # Construct secret credential in the form XXX-XXXX-XXX-XXXX-NNNN
        c = f"{prefix[:3]}-{prefix[3:7]}-{prefix[7:10]}-{prefix[10:14]}-{index:04}"
        return c, prefix, index

    @staticmethod
    def generate_voters_credentials_and_salts(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None, None
        private_credentials = {}
        public_credentials = {}
        salts = {}
        index_already_used = []
        for i, voter in enumerate(election_bundle.voters):
            c, prefix, index = CredentialAuthorityCore.generate_secret_credential(index_already_used)
            private_credential = PrivateCredential(index, c, voter.email)

            index_already_used.append(index)

            s = Utility.generate_base58_string(22)
            for i in range(10000):
                if index in index_already_used:
                    s = Utility.generate_base58_string(22)
                else:
                    break
            salt = SaltModel()
            salt.index = index
            salt.salt_value = s
            salt.election_bundle = election_bundle
            DbSession().session.add(salt)

            x = Utility.derive_secret_exponent(prefix, salt, election_bundle.election)
            public_key = pow(election_bundle.election.group.g, x,
                             election_bundle.election.group.p)  # Calculate g^x mod p

            credential = CredentialModel()
            credential.index = index
            credential.public_credential = public_key
            credential.weight = voter.weight
            credential.email= voter.email
            credential.election = election_bundle.election
            election_bundle.public_credentials.append(credential)

            # voter_credential = VoterCredentialModel()
            # voter_credential.election = election_bundle.election
            # voter_credential.voter = voter
            # voter_credential.credential = credential
            # DbSession().session.add(voter_credential)


            DbSession().session.commit()

            private_credentials[index] = asdict(private_credential)
            public_credentials[index] = credential.to_json()
            salts[index] = salt.to_json()

        return private_credentials, public_credentials, salts

    @staticmethod
    def format_output_voters_credentials_and_salts(credentials, salts):
        pass

    @staticmethod
    def send_to_voters_their_credentials(app, election_bundle, private_credentials):
        for voter_credential_id, c in private_credentials.items():
            voter = DbSession().session.query(VoterModel).filter_by(id=voter_credential_id).first()

            print("Sending mail to voter: %s" % (voter.email))
            Utility.send_mail(app, election_bundle.election.credential_authority, voter.email,
                              "You're credential is : %s" % c)



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.GType import GType
from belenios.models.BaseModel import BaseModel
from belenios.models.belenios.EncryptedMsgModel import EncryptedMsgModel


class ChannelMsgModel(BaseModel):
    recipient = Column(GType, nullable=False)
    message_id = Column(Integer, ForeignKey('encrypted_msg_model.id'), nullable=False)
    message: Mapped[EncryptedMsgModel] = relationship('EncryptedMsgModel', uselist=False)
    @staticmethod
    def load_from_json( dict):
        return ChannelMsgModel(recipient=dict['recipient'], message=EncryptedMsgModel.load_from_json(dict['message']))



from Crypto.Util.number import inverse, bytes_to_long, long_to_bytes

from belenios.config.config import AppConfig
from belenios.dataclass.dataclass import KeysPair, EncryptedText, EncryptedTextAndK
from belenios.utilities.Utility import Utility


class ElGamal:
    def discrete_logarithm(self, g_pow_v, g, q, max_dlp=None):
        if q.bit_length() > 64 and max_dlp is not None:
            m = max_dlp
        else:
            if max_dlp is None:
                m = int(q ** 0.5) + 1
            else:
                m = min(int(q ** 0.5) + 1, max_dlp)
        n = 10000
        start = 0
        giant_steps = {}
        giant = 1
        while start < m:
            end = min(start + n, m)
            # print((start, end))
            res, giant_step = self.discrete_logarithm_sub(g_pow_v, g, q, start, end, giant_steps, giant)
            if res is not None:
                return res
            start = end

        return None  # No logarithm found within the range

    def discrete_logarithm_sub(self, g_pow_v, g, p, m_min, m_max, giant_steps, giant):
        # Precompute giant steps
        for j in range(m_min, m_max):
            giant_steps[giant] = j
            giant = (giant * g) % p

        # Compute baby steps
        inv_g_m = pow(inverse(g, p), m_max, p)
        baby = g_pow_v
        for i in range(m_min, m_max):
            if baby in giant_steps:
                return i * m_max + giant_steps[baby], None
            baby = (baby * inv_g_m) % p

        return None, giant_steps  # No logarithm found within the range

    def split_into_chunks(self, number, p):
        """
        Split a large number into smaller chunks.

        Args:
            number (int): The input number to split.

        Returns:
            list: A list of smaller integers.
        """
        # Determine chunk size based on the size of p
        chunk_size = (p - 1).bit_length() - 1  # Using (p - 1) to ensure the chunks are smaller than p

        chunks = []
        first = True
        while number or first:
            # Extract the least significant chunk_size bits
            chunk = number % (1 << chunk_size)
            chunks.append(chunk)
            # Shift number to the right by chunk_size bits
            number >>= chunk_size
            first = False
        return chunks[::-1]

    def recover_number(self, chunks, p):
        """
        Recover the original number from smaller chunks.

        Args:
            chunks (list): List of smaller integers.
            chunk_size (int): The size of each chunk.

        Returns:
            int: The original number.
        """
        # Determine chunk size based on the size of p
        chunk_size = (p - 1).bit_length() - 1  # Using (p - 1) to ensure the chunks are smaller than p
        number = 0
        for chunk in chunks:
            if chunk is None:
                return None
            number = (number << chunk_size) + chunk
        return number

    # Asymmetric encryption
    def encrypt(self, msg, public_key, group, save_k=False):
        # Convert the plaintext to bytes and then to an integer
        converted_number = bytes_to_long(msg)
        return self.encrypt_from_int(converted_number,public_key, group, save_k)
    # Asymmetric encryption
    def encrypt_from_int(self, number, public_key, group, save_k=False, exponential=False):
        en_msg = []
        # Split the number into smaller chunks
        chunks = self.split_into_chunks(number, group.q)
        # print("Chunks:", chunks)
        for i in range(0, len(chunks)):
            k = Utility.generate_number_in_q(group)  # Random key
            k = 10
            u = pow(group.g, k, group.p)
            if exponential is True:
                v = (pow(public_key, k, group.p) * pow(group.g, number, group.p)) % group.p
                # v = (pow(public_key, k, group.p) * pow(group.g, number, group.p))
            else:
                v = (pow(public_key, k, group.p) * number) % group.p

            # print("g^k (u) used : ", u)
            # print("g^ak (v) used : ", v)
            if save_k:
                en_msg.append((u, v, k))
            else:
                en_msg.append((u, v))

        if save_k:
            return EncryptedTextAndK(en_msg)
        return EncryptedText(en_msg)

    def decrypt_from_int(self, encrypted_text, private_key, group, exponential=False, right=None,
                         no_dlp_searching=False, max_dlp=None):
        dr_msg = []
        for i in range(0, len(encrypted_text.chunk)):
            chunk = encrypted_text.chunk[i]
            u = chunk[0]
            v = chunk[1]
            if right is None:
                right = pow(u, -1 * private_key, group.p)
            if exponential is True:
                # print("exp")
                m_prime = (v * right) % group.p
                # m_prime = g^m . Find m for known m_prime and known g (DLP).
                # return bsgs( group.g,m_prime, group.q)
                if no_dlp_searching:
                    d = m_prime
                else:
                    d = self.discrete_logarithm(m_prime, group.g, group.p, max_dlp)
            else:
                d = (v * right) % group.p
            dr_msg.append(d)

        # print("dr_msg:", dr_msg)
        recovered_number = self.recover_number(dr_msg, group.q)
        # print("Recovered number:", recovered_number)
        return recovered_number

    def decrypt(self, encrypted_text, private_key, group, exponential=False):
        recovered_number = self.decrypt_from_int(encrypted_text, private_key, group, exponential)
        regenerated_string = long_to_bytes(recovered_number).decode()
        return regenerated_string


    # Driver code
    def main(self):
        # p = 11
        # g = 3
        # q = 5
        # group = AppConfig().el_gamal_groups.get("just-for-test")
        group = AppConfig().el_gamal_groups.get("BELENIOS-2048")
        # Bob key generation
        # Generate a private key
        # x = getRandomRange(1, group.q - 2)
        x = 5
        y = pow(group.g, x, group.p)
        keys_pair = KeysPair(y, x)
        # print("g used : ", group.g)
        # print("g^a (public_key) used : ", keys_pair.public_key)

        # Alice's encryption message with Bob public key
        # encrypted_text = self.encrypt(msg.encode(), keys_pair.public_key, group)
        # exponential = False
        exponential = True
        # exit(False)
        encrypted_text1 = self.encrypt_from_int(10, keys_pair.public_key, group, save_k=False, exponential=exponential)
        decrypted_text = self.decrypt_from_int(encrypted_text1, keys_pair.private_key, group, exponential=exponential,
                                               max_dlp=10 ** 2)
        print(decrypted_text)
        if decrypted_text != 10:
            exit(False)

        encrypted_text2 = self.encrypt_from_int(10, keys_pair.public_key, group, save_k=False, exponential=exponential)
        decrypted_text = self.decrypt_from_int(encrypted_text2, keys_pair.private_key, group, exponential=exponential,
                                               max_dlp=10 ** 2)
        print(decrypted_text)
        if decrypted_text != 10:
            exit(False)
        # exit()

        f = EncryptedText([((encrypted_text1.chunk[0][0] * encrypted_text2.chunk[0][0]) % group.p,
                            (encrypted_text1.chunk[0][1] * encrypted_text2.chunk[0][1]) % group.p)])
        # f = EncryptedText([(pow(encrypted_text1.chunk[0][0],2,group.p) , pow(encrypted_text1.chunk[0][1],2,group.p))])
        # print(f)
        # exit()
        decrypted_text = self.decrypt_from_int(f, keys_pair.private_key, group, exponential=exponential,
                                               max_dlp=10 ** 5)
        print(decrypted_text)
        if decrypted_text != 20:
            exit(False)

        encrypted_text1 = self.encrypt_from_int(10 ** 5, keys_pair.public_key, group, save_k=False,
                                                exponential=exponential)
        decrypted_text = self.decrypt_from_int(encrypted_text1, keys_pair.private_key, group, exponential=exponential,
                                               max_dlp=10 ** 6)
        print(decrypted_text)
        if decrypted_text != 10 ** 5:
            exit(False)

        encrypted_text2 = self.encrypt_from_int(10, keys_pair.public_key, group, save_k=False,
                                                exponential=exponential)
        decrypted_text = self.decrypt_from_int(encrypted_text2, keys_pair.private_key, group, exponential=exponential,
                                               max_dlp=10 ** 3)
        print(decrypted_text)
        if decrypted_text != 10:
            exit(False)
        # exit()

        f = EncryptedText([((encrypted_text1.chunk[0][0] * encrypted_text2.chunk[0][0]) % group.p,
                            (encrypted_text1.chunk[0][1] * encrypted_text2.chunk[0][1]) % group.p)])
        # f = EncryptedText([(pow(encrypted_text1.chunk[0][0],2,group.p) , pow(encrypted_text1.chunk[0][1],2,group.p))])
        # print(f)
        # exit()
        decrypted_text = self.decrypt_from_int(f, keys_pair.private_key, group, exponential=exponential,
                                               max_dlp=10 ** 9)
        print(decrypted_text)
        if decrypted_text != (10 ** 5) + (10):
            exit(False)
        exit(True)
        decrypted_text = self.decrypt_from_int(f, keys_pair.private_key, group, exponential=True, max_dlp=10 ** 5)
        print("Decrypted Message :", decrypted_text)
        exit()
        if decrypted_text == 50000 + 604321:
            print("is homomorphic by sum")
        decrypted_text = self.decrypt_from_int(f, keys_pair.private_key, group, exponential=True, max_dlp=10 ** 5)
        print("Decrypted Message :", decrypted_text)
        exit()
        encrypted_text1 = self.encrypt_from_int(10, keys_pair.public_key, group, save_k=False, exponential=False)
        # print(encrypted_text1)
        encrypted_text2 = self.encrypt_from_int(10, keys_pair.public_key, group, save_k=False, exponential=False)
        # print(encrypted_text2)

        f = EncryptedText([(encrypted_text1.chunk[0][0] * encrypted_text2.chunk[0][0],
                            encrypted_text1.chunk[0][1] * encrypted_text2.chunk[0][1])])
        decrypted_text = self.decrypt_from_int(f, keys_pair.private_key, group, exponential=False)

        if decrypted_text == 10 * 10:
            print("is homomorphic by mul")
        print("Decrypted Message :", decrypted_text)
        # print("Is Message correct :", msg == decrypted_text)

        # a = EncryptedTextAndK(chunk=[(4668496126002962658882380217764192095482229044353957117271293358702925124310,
        #                               30397351926047547824654525111658362281056689471750309174167068820938261670430,
        #                               49164449440762570282379174637271973724871893072529699739403694160311175459599)])
        # pk = 46848588746177681989295036151738831494583428455369152762249665817922661702588
        # pub = 66003342716164202112599502859642890596263702302645409972366290560489068378251
        # pub2 = pow(group.g, pk, group.q)
        # print("pub", pub == pub2, pub2)
        # right = 1842561303388720141086967684726446539599603118724140243523399057502308417786670020647042626062945533525282363258206419086505960146180970575368136477223948706063675097215028411798029050315690578296129426861953639126723525644368988519151481189078378712844441981296086127497349204213478543795633586046613583090647511436030387362640850775912085133649344927468703990258714127783000273033922947006372940887198361548874714524620270794156832114456759242638903314506860253621014015049420318494360289857435405025449163951034744450081017573083116399442458111588827827127402104349888893533522797993830704967457983704703915756214
        # # a = ElGamal().encrypt_from_int(1, pub, group, save_k=True, exponential=True)
        # print("dsf", self.decrypt_from_int(a, pk, group, exponential=True, right=right))

if __name__ == '__main__':
    ElGamal().main()
